from textwrap import dedent

import typer

from nef_pipelines.lib.structures import (
    AtomLabel,
    SequenceResidue,
    ShiftData,
    ShiftList,
)
from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.nmrpipe.importers.shifts import shifts as shifts_app
from nef_pipelines.transcoders.nmrpipe.nmrpipe_lib import (
    read_db_file_records,
    read_shift_file,
)

app = typer.Typer()
app.command()(shifts_app)

SHIFTS_NMRPIPE = "nef_chemical_shift_list_nmrpipe"

HEADER = open(path_in_test_data(__file__, "test_header_entry.txt")).read()


def test_lib_parse_shifts():

    EXPECTED = ShiftList(
        shifts=[
            ShiftData(
                atom=AtomLabel(
                    SequenceResidue(
                        chain_code="A", sequence_code=1, residue_name="ALA"
                    ),
                    atom_name="N",
                ),
                value=125.74,
                value_uncertainty=None,
            ),
            ShiftData(
                atom=AtomLabel(
                    SequenceResidue(
                        chain_code="A", sequence_code=2, residue_name="GLY"
                    ),
                    atom_name="HN",
                ),
                value=8.31,
                value_uncertainty=None,
            ),
        ]
    )

    test_data = """
        VARS   RESID RESNAME ATOMNAME SHIFT
        FORMAT %4d %1s %4s %8.2f

        1   ALA   N   125.74
        2   GLY   HN   8.31

    """

    test_data = dedent(test_data)

    test_records = read_db_file_records(test_data.split("\n"))

    shifts = read_shift_file(test_records)
    assert shifts == EXPECTED


# noinspection PyUnusedLocal
def test_ns3_S135A_BMRB1_short():
    EXPECTED = """\
        save_nef_chemical_shift_list_nmrpipe
           _nef_chemical_shift_list.sf_category   nef_chemical_shift_list
           _nef_chemical_shift_list.sf_framecode  nef_chemical_shift_list_nmrpipe

           loop_
              _nef_chemical_shift.chain_code
              _nef_chemical_shift.sequence_code
              _nef_chemical_shift.residue_name
              _nef_chemical_shift.atom_name
              _nef_chemical_shift.value
              _nef_chemical_shift.value_uncertainty
              _nef_chemical_shift.element
              _nef_chemical_shift.isotope_number

             A   1   ALA   N     125.74   .   .   .
             A   1   ALA   C     179.08   .   .   .
             A   2   GLY   CA    44.77    .   .   .
             A   2   GLY   HA2   3.89     .   .   .
             A   2   GLY   HA3   3.8      .   .   .

           stop_

        save_
    """
    path = path_in_test_data(__file__, "ns3_S135A_BMRB1_short.txt")
    result = run_and_report(app, [path], input=HEADER)

    mol_sys_result = isolate_frame(result.stdout, SHIFTS_NMRPIPE)

    assert_lines_match(EXPECTED, mol_sys_result)


def test_residue_1let_3let_translation():
    EXPECTED = """\
        save_nef_chemical_shift_list_nmrpipe
            _nef_chemical_shift_list.sf_category   nef_chemical_shift_list
            _nef_chemical_shift_list.sf_framecode  nef_chemical_shift_list_nmrpipe

            loop_
                _nef_chemical_shift.chain_code
                _nef_chemical_shift.sequence_code
                _nef_chemical_shift.residue_name
                _nef_chemical_shift.atom_name
                _nef_chemical_shift.value
                _nef_chemical_shift.value_uncertainty
                _nef_chemical_shift.element
                _nef_chemical_shift.isotope_number

                A   1   MET   CB   33.27    .   .   .
                A   1   MET   CA   54.45    .   .   .
                A   1   MET   C    170.54   .   .   .
                A   1   MET   HA   4.23     .   .   .
                A   2   GLN   HN   8.9      .   .   .
                A   2   GLN   CA   55.08    .   .   .
                A   2   GLN   CB   30.76    .   .   .
                A   2   GLN   C    175.92   .   .   .
                A   2   GLN   HA   5.249    .   .   .
                A   2   GLN   N    123.22   .   .   .

           stop_

        save_
    """

    path = path_in_test_data(__file__, "P3a_l273R_nmrpipe_shifts_short.tab")

    result = run_and_report(app, [path], input=HEADER)

    mol_sys_result = isolate_frame(result.stdout, SHIFTS_NMRPIPE)

    assert_lines_match(EXPECTED, mol_sys_result)



from random import seed

import pytest


@pytest.fixture
def fixed_seed():
    seed(42)


@pytest.fixture
def clear_cache():
    """
    clear the lru cache used by lib.util cached_stdin
    see https://stackoverflow.com/questions/40273767/clear-all-lru-cache-in-python
    """
    import functools
    import gc

    gc.collect()
    wrappers = []
    for elem in gc.get_objects():
        try:
            if isinstance(elem, functools._lru_cache_wrapper):
                wrappers.append(elem)
        except ReferenceError:
            pass

    for wrapper in wrappers:
        wrapper.cache_clear()


def pytest_configure(config):
    from nef_pipelines.main import create_nef_app

    create_nef_app()



import typer

from nef_pipelines import nef_app

peaks_app = typer.Typer()


if nef_app.app:
    nef_app.app.add_typer(
        peaks_app,
        name="peaks",
        help="-  carry out operations on nef peaks",
    )

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.tools.peaks.match  # noqa: F401



# TODO add support for DATA FIRST_RESID
# noinspection PyUnusedLocal
from pathlib import Path
from typing import Iterable, List

import typer
from pynmrstar import Entry

from nef_pipelines.lib.nef_lib import (
    add_frames_to_entry,
    read_file_or_exit,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.sequence_lib import (
    get_chain_starts,
    offset_chain_residues,
    sequence_to_nef_frame,
)
from nef_pipelines.lib.structures import SequenceResidue
from nef_pipelines.lib.util import exit_error
from nef_pipelines.transcoders.nmrpipe import import_app
from nef_pipelines.transcoders.nmrpipe.nmrpipe_lib import (
    gdb_to_sequence,
    read_db_file_records,
)

app = typer.Typer()


# noinspection PyUnusedLocal
@import_app.command()
def sequence(
    chain_code: str = typer.Option(
        "A",
        "--chains",
        help="the chain code to use",
        metavar="<CHAIN-CODE>",
    ),
    no_chain_start: bool = typer.Option(
        False,
        "--no-chain-start",
        help="don't include a start of chain link type for the first residue",
    ),
    no_chain_end: bool = typer.Option(
        False,
        "--no-chain-end",
        help="don't include an end of chain link type for the last residue",
    ),
    entry_name: str = typer.Option(
        "nmrpipe", help="a name for the entry if its created from new"
    ),
    input: Path = typer.Option(
        None,
        "-i",
        "--input",
        metavar="|INPUT|",
        help="input to read NEF data from",
    ),
    start: int = typer.Option(
        None, metavar="<START>", help="residue number to start the sequence at"
    ),
    file_name: Path = typer.Argument(
        ..., help="input files of type <NMRPIPE>.tab", metavar="<NMRPIPE-FILE>"
    ),
):
    """read a sequence from an NMRPipe file <NMPRPIPE-FILE>.tab into NEF a file"""

    entry = read_or_create_entry_exit_error_on_bad_file(input, entry_name=entry_name)

    lines = read_file_or_exit(file_name)

    entry = pipe(
        entry, lines, chain_code, no_chain_start, no_chain_end, start, file_name
    )

    print(entry)


def pipe(
    entry: Entry,
    lines: List[str],
    chain_code: str,
    no_chain_start: bool,
    no_chain_end: bool,
    start: int,
    file_name: str,
):
    try:

        nmrpipe_sequence = read_sequence(
            lines,
            chain_code=chain_code,
            sequence_file_name=file_name,
            start=start,
        )

        no_chain_start_list = (
            [
                chain_code,
            ]
            if no_chain_start is True
            else []
        )
        no_chain_end_list = (
            [
                chain_code,
            ]
            if no_chain_end is True
            else []
        )

        frame = sequence_to_nef_frame(
            nmrpipe_sequence,
            no_chain_start=no_chain_start_list,
            no_chain_end=no_chain_end_list,
        )

        entry = add_frames_to_entry(
            entry,
            [
                frame,
            ],
        )

    except Exception as e:
        exit_error(f"reading sequence failed because {e}", e)

    return entry


def read_sequence(
    sequence_lines: Iterable[str],
    chain_code: str = "A",
    sequence_file_name: str = "unknown",
    start: int = None,
) -> List[SequenceResidue]:

    gdb_file = read_db_file_records(sequence_lines, sequence_file_name)

    sequence = gdb_to_sequence(gdb_file, chain_code)

    if sequence and start is not None:
        nmrpipe_chain_start = get_chain_starts(sequence)[chain_code]

        offset = start - nmrpipe_chain_start
        sequence = offset_chain_residues(sequence, {chain_code: offset})

    return sequence


if __name__ == "__main__":

    typer.run(sequence)



# original implementation by esther
import sys
from pathlib import Path

import typer

from nef_pipelines.lib.nef_lib import read_entry_from_stdin_or_exit
from nef_pipelines.transcoders.mars import export_app

MARS_DEFAULT_FILE = "mars.inp"
STDOUT_PATH = "-"


def convert_output_to(file_name, end, extension):
    if file_name != "-":
        path = Path(file_name)
        suffixes = "".join(path.suffixes)
        result = f"{file_name[:-len(suffixes)]}{end}.{extension}"
    else:
        result = file_name

    return result


# noinspection PyUnusedLocal
@export_app.command()
def input(
    root: str = typer.Option(
        None, help="root name for files [default: <ENTRY-ID>]", metavar="<ROOT>"
    ),
    deuterated: bool = typer.Option(
        False,
        help="is the protein deuterated [default: False]",
        metavar="<DEUTERATED-PROTEIN>",
    ),
    folded: bool = typer.Option(
        True, help="is the protein folded [default: False]", metavar="<FOLDED-PROTEIN>"
    ),
    output_file: str = typer.Argument(
        None,
        help="file name to output to [default <entry_id>.inp] for stdout use -",
        metavar="<MARS-INPUT-FILENAME>",
    ),
):
    """- write  mars input fixed assignment and connectivity files"""

    entry = read_entry_from_stdin_or_exit()

    output_file = f"{entry.entry_id}_mars.inp" if output_file is None else output_file
    pred_file = (
        f"{entry.entry_id}_pred.tab"
        if output_file is None
        else convert_output_to(output_file, "_pred", "tab")
    )
    ass_file = (
        f"{entry.entry_id}_fix_ass.tab"
        if output_file is None
        else convert_output_to(output_file, "_fix_ass", "tab")
    )

    deuterated = int(deuterated)
    random_coil = int(not folded)

    root = entry.entry_id if root is None else root

    template = f"""\
        fragSize:     5               # Maximum length of pseudoresidue fragments

        cutoffCO:     0.25            # Connectivity cutoff (ppm) of CO [0.25]
        cutoffCA:     0.2             # Connectivity cutoff (ppm) of CA [0.5]
        cutoffCB:     0.5             # Connectivity cutoff (ppm) of CB [0.5]
        cutoffHA:     0.25            # Connectivity cutoff (ppm) of HA [0.25]
        cutoffHN:     0.02            # Connectivity cutoff (ppm) of HN [0.15]
        cutoffN:      0.05            # Connectivity cutoff (ppm) of N [0.10]
        cutoffnHN:    0.02            # Connectivity cutoff (ppm) of HN+1 [0.15]
        cutoffnN:     0.05            # Connectivity cutoff (ppm) of N+1 [0.10]

        fixConn:      {root}_fix_con.tab     # Table for fixing sequential connectivity
        fixAss:       {root}_fix_ass.tab     # Table for fixing residue type and(or) assignment

        pdb:          0               # 3D structure available [0/1]
        resolution:   NO              # Resolution of 3D structure [Angstrom]
        pdbName:      NO              # Name of PDB file (protons required!)
        tensor:       NO              # Method for obtaining alignment tensor [0/1/2/3/4]

        nIter:        NO              # Number of iterations [2/3/4]

        dObsExh:      NO              # Name of RDC table for exhaustive SVD (PALES format)
        dcTab:        NO              # Name of RDC table (PALES format)

        deuterated:   {deuterated}               # Protonated proteins [0]; perdeuterated proteins [1]

        rand_coil:    {random_coil}              # Folded proteins [0]; disordered proteins [1]

        sequence:     {root}.fasta       # Primary sequence (FASTA format)
        secondary:    {root}_psipred.tab # Secondary structure (PSIPRED format)
        csTab:        {root}_shifts.tab  # Chemical shift table
    """

    file_h = sys.stdout if output_file == STDOUT_PATH else open(output_file, "w")

    print(template, file=file_h)

    if output_file != STDOUT_PATH:
        print("NOTE: writing empty pred and ass files!", file=sys.stderr)
        print(f"    pred file: {pred_file}", file=sys.stderr)
        print(f"    ass file: {ass_file}", file=sys.stderr)

        with open(pred_file, "w") as fh:
            print(file=fh)

        with open(ass_file, "w") as fh:
            print(file=fh)

    if output_file != STDOUT_PATH:
        file_h.close()

        if not sys.stdout.isatty():
            print(entry)



# TODO: what do unassigned nmrpipe shifts look like?
# TODO: support general nmrpipe shift input
# TODO: support the HA2|HA3 syntax
# TODO: support unassigned

import sys
from enum import auto
from pathlib import Path
from typing import Dict, List, Tuple

import typer
from fyeah import f
from pynmrstar import Loop, Saveframe
from strenum import StrEnum

from nef_pipelines.lib.nef_lib import (
    UNUSED,
    add_frames_to_entry,
    read_file_or_exit,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.sequence_lib import (
    get_residue_name_from_lookup,
    offset_chain_residues,
    sequence_3let_to_sequence_residues,
    sequence_from_entry,
    sequence_to_chains,
    sequence_to_residue_name_lookup,
)
from nef_pipelines.lib.structures import (
    AtomLabel,
    DihedralRestraint,
    Residue,
    SequenceResidue,
)
from nef_pipelines.lib.util import (
    STDIN,
    exit_error,
    is_float,
    parse_comma_separated_options,
)
from nef_pipelines.transcoders.nmrpipe.nmrpipe_lib import (
    VALUES,
    DbFile,
    DbRecord,
    gdb_to_3let_sequence,
    get_column_indices,
    get_gdb_columns,
    read_db_file_records,
    select_data_records,
    select_records,
)
from nef_pipelines.transcoders.talos import import_app
from nef_pipelines.transcoders.talos.talos_lib import _exit_if_sequences_dont_match

app = typer.Typer()

CLASS_TO_MERIT = {
    "None": 0.0,
    "Strong": 1.0,
    "Generous": 0.6,
    "Warn": 0.3,
    "Dyn": 0.0,
    "Bad": 0.0,
}

CLASS_HELP = ",".join([f"{class_},{merit}" for class_, merit in CLASS_TO_MERIT.items()])
MERIT_HELP = f"""
a comma separated list of merit pairs of class (one of Dyn, Strong, Generous Warn Bad or None) and value (a floating
point number) e.g dyn,0.0,strong,7.0 [default is {CLASS_HELP}] multiple values can also be added by multiple uses of the
option
"""


# noinspection PyUnusedLocal
@import_app.command()
def restraints(
    chain_code: str = typer.Option(
        "A",
        "-c",
        "--chain",
        help="chain to import to",
        metavar="<CHAIN-CODE>",
    ),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--input",
        help="input to read NEF data from [- is stdin]",
    ),
    frame_name: str = typer.Option(
        "talos_restraints_{chain_code}_{angle}",
        "-f",
        "--frame",
        help="name for the frame that will be created default is talos_restraints_<CHAIN-CODE>",
    ),
    merits: List[str] = typer.Option([], help=MERIT_HELP),
    file_name: Path = typer.Argument(..., help="input talos output file pred.tab"),
):
    """- convert talos restraints to nef [alpha]"""

    class_to_merit = _parse_merits_and_merge(merits)

    entry = read_or_create_entry_exit_error_on_bad_file(input)

    lines = read_file_or_exit(file_name)

    entry = pipe(entry, lines, file_name, chain_code, frame_name, class_to_merit)

    print(entry)


def _parse_merits_and_merge(merits):
    new_merits = parse_comma_separated_options(merits)

    _exit_error_if_merits_arent_pairs(new_merits)

    new_merits_dict = dict(zip(new_merits[::2], new_merits[1::2]))

    _exit_error_if_merits_have_bad_classes(new_merits_dict)

    new_merits_non_float = {
        key: value for key, value in new_merits_dict.items() if not is_float(value)
    }

    _exit_error_if_merits_arent_floats(new_merits_non_float)

    new_merits = {key.lower(): float(value) for key, value in new_merits_dict.items()}

    return CLASS_TO_MERIT.update(new_merits)


def _exit_error_if_merits_arent_floats(new_merits_non_float):
    if len(new_merits_non_float) > 0:
        new_merits_non_float_string = ", ".join(
            [f"{key}:{value}" for key, value in new_merits_non_float.items()]
        )
        msg = f"""
            merit values must be floats, the following class,values pairs have non float values

            {new_merits_non_float_string}

        """
        exit_error(msg)


def _exit_error_if_merits_have_bad_classes(new_merits_dict):
    unexpected_classes = set(new_merits_dict) - set(CLASS_TO_MERIT.keys())
    if len(unexpected_classes) > 0:
        talos_classes_string = ", ".join(CLASS_TO_MERIT.keys())
        new_class_string = ", ".join(unexpected_classes)
        msg = f"""
        some merit classes provided are not used by TALOS

        Talos classes are:

        {talos_classes_string}

        unexpected classes are:

        {new_class_string}

        """
        exit_error(msg)


def _exit_error_if_merits_arent_pairs(new_merits):
    if len(new_merits) % 2 != 0:
        merit_string = ", ".join(new_merits)
        msg = f"""
        merit values must come in pairs I got {len(new_merits)} values:

        {merit_string}
        """

        exit_error(msg)


DIHEDRAL_FRAME_CATEGORY = "nef_dihedral_restraint_list"
DIHEDRAL_RESTRAINT_TAGS = """index restraint_id restraint_combination_id name
                                chain_code_1 sequence_code_1 residue_name_1 atom_name_1
                                chain_code_2 sequence_code_2 residue_name_2 atom_name_2
                                chain_code_3 sequence_code_3 residue_name_3 atom_name_3
                                chain_code_4 sequence_code_4 residue_name_4 atom_name_4
                                target_value target_value_error
                                ccpn_comment np_merit""".split()


def _dihedral_restraints_to_frame(dihedral_restraints, frame_code, comment):

    frame_code = f"{DIHEDRAL_FRAME_CATEGORY}_{frame_code}"

    frame = Saveframe.from_scratch(frame_code, DIHEDRAL_FRAME_CATEGORY)
    frame.add_tag("sf_category", DIHEDRAL_FRAME_CATEGORY)
    frame.add_tag("sf_framecode", frame_code)
    frame.add_tag("potential_type", "square-well-parabolic")
    frame.add_tag("ccpn_comment", comment)

    loop = Loop.from_scratch("nef_dihedral_restraint")

    for tag in DIHEDRAL_RESTRAINT_TAGS:
        loop.add_tag(tag)

    for index, dihedral_restraint in enumerate(dihedral_restraints, start=1):
        data = {
            "index": index,
            "restraint_id": index,
            "restraint_combination_id": UNUSED,
            "name": dihedral_restraint.name,
            "chain_code_1": dihedral_restraint.atom_1.residue.chain_code,
            "sequence_code_1": dihedral_restraint.atom_1.residue.sequence_code,
            "residue_name_1": dihedral_restraint.atom_1.residue.residue_name,
            "atom_name_1": dihedral_restraint.atom_1.atom_name,
            "chain_code_2": dihedral_restraint.atom_2.residue.chain_code,
            "sequence_code_2": dihedral_restraint.atom_2.residue.sequence_code,
            "residue_name_2": dihedral_restraint.atom_2.residue.residue_name,
            "atom_name_2": dihedral_restraint.atom_2.atom_name,
            "chain_code_3": dihedral_restraint.atom_3.residue.chain_code,
            "sequence_code_3": dihedral_restraint.atom_3.residue.sequence_code,
            "residue_name_3": dihedral_restraint.atom_3.residue.residue_name,
            "atom_name_3": dihedral_restraint.atom_3.atom_name,
            "chain_code_4": dihedral_restraint.atom_4.residue.chain_code,
            "sequence_code_4": dihedral_restraint.atom_4.residue.sequence_code,
            "residue_name_4": dihedral_restraint.atom_4.residue.residue_name,
            "atom_name_4": dihedral_restraint.atom_4.atom_name,
            "target_value": round(dihedral_restraint.target_value, 3),
            "target_value_error": round(dihedral_restraint.target_value_error, 3),
            "np_merit": dihedral_restraint.merit,
            "ccpn_comment": dihedral_restraint.remark,
        }

        loop.add_data(
            [
                data,
            ]
        )

    frame.add_loop(loop)

    return frame


def _get_source_file(gdb_records: DbFile):
    def predicate(x):
        return "Prediction Summary for Chemical Shift Input" in x.values

    records = select_records(gdb_records, "REMARK", predicate)

    file = "unknown"

    if len(records) == 1:
        file = records[0].values.split()[-1]

    return file


def _get_talos_version(gdb_records: DbFile):
    def predicate(x):
        return "Version" in x.values and "INFO" in x.values

    records = select_records(gdb_records, "REMARK", predicate)

    version = "unknown"

    if len(records) == 1:
        version = " ".join(records[0].values.split()[3:-1])

    return version


def pipe(entry, lines, file_name, chain_code, frame_name, class_to_merit=None):

    class_to_merit = CLASS_TO_MERIT if class_to_merit is None else class_to_merit

    nef_sequence = sequence_from_entry(entry)

    nef_chain_codes = sequence_to_chains(nef_sequence)

    gdb_records = read_db_file_records(lines)

    talos_restraints, talos_sequence, angle = _read_dihedral_restraints(
        gdb_records,
        chain_code=chain_code,
        file_name=file_name,
        class_to_merit=class_to_merit,
        nef_sequence=nef_sequence,
    )

    frame_name = f(frame_name)

    if chain_code in nef_chain_codes:
        _exit_if_sequences_dont_match(chain_code, nef_sequence, talos_sequence)
    else:
        _exit_error_no_chain_code(chain_code, file_name)

    source_file = _get_source_file(gdb_records)
    talos_version = _get_talos_version(gdb_records)

    comment = (
        f"file: {file_name}, source file: {source_file}, talos version: {talos_version}"
    )

    dihedral_restraint_frame = _dihedral_restraints_to_frame(
        talos_restraints, frame_name, comment
    )

    add_frames_to_entry(
        entry,
        [
            dihedral_restraint_frame,
        ],
    )

    return entry


def _exit_error_no_chain_code(chain_code, file_name):
    args = " ".join(sys.argv[1:])
    msg = f"""
            There is no sequence defined for the chain {chain_code}, please import a sequence! You may want to do
            nef talos import sequence {file_name} | nef {args}
        """
    exit_error(msg)


REQUIRED_COLUMNS_PHI_PSI = "RESID RESNAME PHI PSI DPHI DPSI CLASS".split()
REQUIRED_COLUMNS_CHI = "RESID RESNAME CHI1 DCHI1".split()


class Angles(StrEnum):
    PHI_PSI = auto()
    CHI1 = auto()


def _check_if_phi_psi_or_chi(columns):
    if "PHI" in columns and "PSI" in columns:
        result = Angles.PHI_PSI
    elif "CHI1" in columns:
        result = Angles.CHI1
    else:
        result = None

    return result


def _read_dihedral_restraints(
    gdb_records: List[DbRecord],
    chain_code: str,
    file_name: str,
    class_to_merit: Dict[str, float],
    nef_sequence: List[SequenceResidue],
) -> List[DihedralRestraint]:

    columns = get_gdb_columns(gdb_records)

    angle_type = _check_if_phi_psi_or_chi(columns)

    _exit_if_no_phi_psi_or_chi_detected(angle_type, columns, file_name)

    if angle_type is Angles.PHI_PSI:
        required_columns = REQUIRED_COLUMNS_PHI_PSI
    elif angle_type is Angles.CHI1:
        required_columns = REQUIRED_COLUMNS_CHI

    _exit_if_required_columns_missing(columns, required_columns, file_name)

    sequence_3let = gdb_to_3let_sequence(gdb_records)

    if len(sequence_3let) > 0:

        chain_start = _first_residue_number(gdb_records)

        sequence = sequence_3let_to_sequence_residues(sequence_3let, chain_code)

        sequence = offset_chain_residues(sequence, {chain_code: chain_start - 1})
    else:
        sequence = nef_sequence

    return (
        _gdb_records_to_torsion_restraints(
            gdb_records, sequence, chain_code, class_to_merit, angle_type
        ),
        sequence,
        angle_type,
    )


def _exit_if_no_phi_psi_or_chi_detected(angle_type, columns, file_name):
    if angle_type is None:
        column_string = ", ".join(columns)
        msg = f"""
            in the file {file_name}
            columns found don't match those for phi/psi or chi1, columns were

            {column_string}
        """
        exit_error(msg)


TALOS_MISSING_VALUE_FLOAT = 9999.0
PHI = "PHI"
PSI = "PSI"
CHI1 = "CHI1"
TORSION_ATOMS = {
    PHI: (("C", 1), ("N", 0), ("CA", 0), ("C", 0)),
    PSI: (("N", 0), ("CA", 0), ("C", 0), ("N", 1)),
    CHI1: (("N", 0), ("CA", 0), ("CB", 0), ("*G", 0)),
}

# these should be looked up in chem comps
CHI1_ATOMS = {
    "ARG": "CG",
    "ASN": "CG",
    "ASP": "CG",
    "CYS": "SG",
    "GLN": "CG",
    "GLU": "CG",
    "HIS": "CG",
    "ILE": "CG1",
    "LEU": "CG",
    "LYS": "CG",
    "MET": "CG",
    "PHE": "CG",
    "PRO": "CG",
    "SER": "OG",
    "THR": "OG1",
    "TRP": "CG1",
    "TYR": "CG",
    "VAL": "CG1",
}


def _residue_to_torsion_atoms(
    chain_code,
    sequence_code: int,
    residue_lookup: Dict[Tuple[str, str], str],
    angle_name: str,
) -> List[AtomLabel]:

    labels = []

    for atom_name, offset in TORSION_ATOMS[angle_name]:

        offset_sequence_code = sequence_code + offset

        residue_name = get_residue_name_from_lookup(
            chain_code, offset_sequence_code, residue_lookup
        )

        if atom_name == "*G":
            atom_name = CHI1_ATOMS[residue_name]

        residue = Residue(chain_code, offset_sequence_code, residue_name)

        labels.append(AtomLabel(residue, atom_name))

    return labels


def _gdb_records_to_torsion_restraints(
    gdb_records, sequence, chain_code, class_to_merit, angle_type
):

    residue_lookup = sequence_to_residue_name_lookup(sequence)

    column_indices = get_column_indices(gdb_records)

    restraints = []
    for record in select_records(gdb_records, VALUES):

        class_ = record.values[column_indices["CLASS"]]

        if class_ in ("None", "na"):
            continue

        if angle_type == Angles.PHI_PSI:

            phi = _build_restraint(
                record,
                "PHI",
                column_indices,
                chain_code,
                residue_lookup,
                class_to_merit,
            )
            psi = _build_restraint(
                record,
                "PSI",
                column_indices,
                chain_code,
                residue_lookup,
                class_to_merit,
            )

            restraints.append(phi)
            restraints.append(psi)
        elif angle_type == Angles.CHI1:
            chi1 = _build_restraint(
                record,
                "CHI1",
                column_indices,
                chain_code,
                residue_lookup,
                class_to_merit,
            )
            restraints.append(chi1)

    return restraints


def _build_restraint(
    record, type, column_indices, chain_code, residue_lookup, class_to_merit
):

    sequence_code = record.values[column_indices["RESID"]]
    class_ = record.values[column_indices["CLASS"]]

    angle = record.values[column_indices[type]]
    delta_angle = record.values[column_indices[f"D{type}"]]
    atoms = _residue_to_torsion_atoms(chain_code, sequence_code, residue_lookup, type)

    merit = 0.0
    remark = ""
    if type in ["PHI", "PSI"]:
        merit = class_to_merit[class_]
        remark = f"class: {class_}"
    elif type == "CHI1":
        merit = 1.0
        Q_Gm = record.values[column_indices["Q_Gm"]]
        Q_Gp = record.values[column_indices["Q_Gp"]]
        Q_T = record.values[column_indices["Q_T"]]

        remark = f"Q_Gm: {Q_Gm}, Q_Gt: {Q_Gp}, Q_T: {Q_T}"

    restraint = DihedralRestraint(
        *atoms,
        target_value=angle,
        target_value_error=delta_angle,
        name=type,
        remark=remark,
        merit=merit,
    )

    return restraint


def _first_residue_number(gdb_records):
    first_residue_records = select_data_records(gdb_records, "FIRST_RESID")
    first_residue = 1
    if len(first_residue_records) == 1:
        first_residue = int(first_residue_records[0].values[1])
    return first_residue


def _exit_if_required_columns_missing(columns, required_columns, file_name):
    column_set = set(columns)

    required_column_set = set(required_columns)

    if not required_column_set.issubset(column_set):
        msg = f"""
            the required column in the file {file_name} are:

            {', '.join(required_columns)}

            found columns are

            {', '.join(columns)}

            missing columns are

            {', '.join(required_column_set - column_set)}
        """

        exit_error(msg)



import copy
from textwrap import dedent
from typing import Any, Dict, List, Tuple, Union

from frozendict import frozendict
from ordered_set import OrderedSet
from pynmrstar import Loop, Saveframe

from nef_pipelines.lib.isotope_lib import GAMMA_RATIOS, Isotope
from nef_pipelines.lib.nef_frames_lib import (
    ABSOLUTE_PEAK_POSITIONS,
    ATOM_NAME,
    ATOM_NAME__DIMENSION_INDEX,
    AXIS_CODE,
    AXIS_UNIT,
    CCPN_COMMENT,
    CCPN_MERIT,
    CHAIN_CODE,
    CHAIN_CODE__DIMENSION_INDEX,
    CHEMICAL_SHIFT_LIST,
    DIMENSION__DIMENSION_INDEX,
    DIMENSION_ID,
    DIMENSION_INDEX,
    DIMENSION_LOOP_TAGS,
    EXPERIMENT_CLASSIFICATION,
    EXPERIMENT_TYPE,
    FOLDING,
    HEIGHT,
    HEIGHT_UNCERTAINTY,
    INDEX,
    IS_ACQUISITION,
    IS_INDIRECT,
    NEF_PPM,
    NUM_DIMENSIONS,
    ONE_BOND,
    PEAK_ID,
    PEAK_LOOP_TAGS,
    POSITION__DIMENSION_INDEX,
    POSITION_UNCERTAINTY__DIMENSION_INDEX,
    RESIDUE_NAME,
    RESIDUE_NAME__DIMENSION_INDEX,
    SEQUENCE_CODE,
    SEQUENCE_CODE__DIMENSION_INDEX,
    SF_CATEGORY,
    SF_FRAMECODE,
    SPECTRAL_WIDTH,
    SPECTROMETER_FREQUENCY,
    SPECTRUM_DIMENSION_LOOP_CATEGORY,
    SPECTRUM_DIMENSION_TRANSFER_LOOP_CATEGORY,
    SPECTRUM_FRAME_CATEGORY,
    SPECTRUM_PEAK_LOOP_CATEGORY,
    TRANSFER_LOOP_TAGS,
    TRANSFER_TYPE,
    VALUE_FIRST_POINT,
    VOLUME,
    VOLUME_UNCERTAINTY,
)
from nef_pipelines.lib.nef_lib import (
    NEF_FALSE,
    NEF_NONE,
    NEF_TRUE,
    UNUSED,
    loop_row_dict_iter,
)
from nef_pipelines.lib.structures import (
    AtomLabel,
    DimensionInfo,
    LineInfo,
    NewPeak,
    Residue,
    ShiftData,
)
from nef_pipelines.lib.util import (
    _row_to_table,
    exit_error,
    is_float,
    is_int,
    remove_duplicates_stable,
    unused_to_empty_string,
    unused_to_none,
)


class BadPositionException(Exception):
    """
    The position of a peak in the frame couldn't be converted to a float
    """

    ...


# noinspection PyUnusedLocal
def proton_frequency_to_axis_frequency(
    spectrometer_frequency: float, isotope: Isotope
) -> float:
    """
    Give a proton frequency in MHz and an isotope return the equivalent spectrometer
    frequency for the isotops
    :param spectrometer_frequency: sspectrometer frequency
    :param isotope: an NMR isotope
    :return: the spectrometer frequency for the isotope
    """
    result = float(spectrometer_frequency) * GAMMA_RATIOS[isotope]
    return result


def peaks_to_sweep_width_and_reference_shift(
    peaks: List[NewPeak], dim: int, multiplier: float = 1.1
) -> Tuple[float, float]:
    """
    Given a set of peaks synthesise a sweep width and reference ppm from the min and max shifts of the peaks
    plus a margin
    :param peaks: A list of peaks
    :param dim: the dimesnion of the peaks to use
    :param multiplier: how much to increase the range of the observed chemical shifts by to get a sweep width
    :return: tuple sweep width [ppm], reference shift[ppm]
    """
    shifts = []
    for peak in peaks:
        shifts.append(peak.shifts[dim].value)

    min_shift = min(shifts)
    max_shift = max(shifts)

    width = max_shift - min_shift
    width_d2 = width / 2.0
    middle = min_shift + width_d2

    width_d2_mult = width_d2 * multiplier
    width_mult = width_d2_mult * 2
    min_shift_mult = middle + width_d2_mult

    return width_mult, min_shift_mult


def format_7_3_f(value: Union[int, float]) -> str:
    """
    format a number with 7.3f formatting as a string
    :param value: an int of a float
    :return: a formatted string
    """
    return f"{value:7.3f}".strip()


def _expand_templates(
    template: List[Union[str, List[str]]], values_list: List[Dict[str, Any]]
) -> List[str]:
    """
    expand a list of templates into strings using a list of sets (dicts) of values
    :param templates: the tags to format
    :param values_list: a list of dictionaries of values
    :return: a list of formatted strings
    """
    result = OrderedSet()
    for tag_or_tag_list in template:
        if isinstance(tag_or_tag_list, str):
            for values in values_list:
                result.add(tag_or_tag_list.format(**values))
        elif isinstance(tag_or_tag_list, list):
            for values in values_list:
                for tag in tag_or_tag_list:
                    result.add(tag.format(**values))

    return list(result)


DEFAULT_EXTRA_TAGS = {
    CHEMICAL_SHIFT_LIST: UNUSED,
    EXPERIMENT_CLASSIFICATION: UNUSED,
    EXPERIMENT_TYPE: UNUSED,
}


def peaks_to_frame(
    peaks: List[NewPeak],
    dimensions: List[DimensionInfo],
    spectrometer_frequency: float,
    frame_code: str = "peaks",
    extra_tags: Dict[str, str] = frozendict(DEFAULT_EXTRA_TAGS),
    source: str = "unknown",
) -> Saveframe:

    # TODO needs a list of transfers
    """
    convert a list of peaks to a nef spectrum frame
    :param peaks: the peaks
    :param dimensions: the dimensions of the peak list
    :param spectrometer_frequency: the 1H spectrometer frequency of the peak list
    :param frame_code: the name for the frame
    :param extra_tags: extra tags to be used in the frame, non nef tags will be appended
    :param source: the source of the data for error reporting
    :return: a NEF Spectrum save frame
    """

    mutable_extra_tags = {**DEFAULT_EXTRA_TAGS}
    mutable_extra_tags.update(extra_tags)
    extra_tags = mutable_extra_tags

    frame_code = f"{SPECTRUM_FRAME_CATEGORY}_{frame_code}"

    frame = Saveframe.from_scratch(frame_code, SPECTRUM_FRAME_CATEGORY)

    frame.add_tag(SF_CATEGORY, SPECTRUM_FRAME_CATEGORY)
    frame.add_tag(SF_FRAMECODE, frame_code)

    have_comments = False
    for peak in peaks:
        if peak.comment != "":
            have_comments = True
            break

    have_merits = False
    for peak in peaks:
        if peak.figure_of_merit != UNUSED and peak.figure_of_merit is not None:
            have_merits = True
            break

    peak_dimensions = []
    for peak in peaks:
        peak_dimensions.append(len(peak.shifts))

    peak_dimensions.sort()
    peak_dimensions = remove_duplicates_stable(peak_dimensions)

    if len(peak_dimensions) > 1:
        msg = f"""\
            unexpected! in the file {source} there are peaks with different numbers of different dimensions
            the dimensions were: {' '.join([str(dimension) for dimension in peak_dimensions])}
        """
        exit_error(msg)

    frame.add_tag(NUM_DIMENSIONS, peak_dimensions[0])

    shift_list_name = (
        extra_tags[CHEMICAL_SHIFT_LIST] if CHEMICAL_SHIFT_LIST in extra_tags else UNUSED
    )
    frame.add_tag(
        CHEMICAL_SHIFT_LIST, shift_list_name
    )  # technically not correct to use unused by default here but this is a nef file building
    # tool kit so an empty shift list maybe what we need
    # we may need to add a flag to build the shift list on the fly at a later date...

    experiment_classification = (
        extra_tags[EXPERIMENT_CLASSIFICATION]
        if EXPERIMENT_CLASSIFICATION in extra_tags
        else UNUSED
    )
    frame.add_tag(
        EXPERIMENT_CLASSIFICATION, experiment_classification
    )  # add via an experiment info structure currently uses extra tags?

    experiment_type = (
        extra_tags[EXPERIMENT_TYPE] if EXPERIMENT_TYPE in extra_tags else UNUSED
    )
    frame.add_tag(EXPERIMENT_TYPE, experiment_type)

    dimension_loop = Loop.from_scratch(SPECTRUM_DIMENSION_LOOP_CATEGORY)
    frame.add_loop(dimension_loop)

    dimension_loop.add_tag(DIMENSION_LOOP_TAGS)

    dimensions = copy.deepcopy(dimensions)
    for i, dimension in enumerate(dimensions):
        dimension[DIMENSION_ID] = i + 1

        dimension[AXIS_UNIT] = NEF_PPM

        axis_code = dimension[AXIS_CODE]

        # TODO: need to check we have an axis code here
        # TODO: need to reasonable translations here H-> 1H H1 -> 1H [especially as 1H isn't possible because of
        #  technical limitations python names can't start with a number]
        dimension_spectrometer_frequency = proton_frequency_to_axis_frequency(
            spectrometer_frequency, axis_code
        )
        dimension[SPECTROMETER_FREQUENCY] = format_7_3_f(
            dimension_spectrometer_frequency
        )

        sweep_width, max_shift = peaks_to_sweep_width_and_reference_shift(peaks, i)

        dimension[SPECTRAL_WIDTH] = format_7_3_f(sweep_width)
        dimension[VALUE_FIRST_POINT] = format_7_3_f(max_shift)
        dimension[FOLDING] = NEF_NONE

        dimension[ABSOLUTE_PEAK_POSITIONS] = NEF_TRUE

        acquisition = NEF_TRUE if i == len(dimensions) - 1 else NEF_FALSE

        dimension[IS_ACQUISITION] = acquisition

        dimension_loop.add_data([dimension])

    transfer_loop = Loop.from_scratch(SPECTRUM_DIMENSION_TRANSFER_LOOP_CATEGORY)
    frame.add_loop(transfer_loop)

    dimension_indices = [
        {"dimension_index": dimension_index}
        for dimension_index in range(1, len(dimensions) + 1)
    ]
    transfer_loop_tags = _expand_templates(TRANSFER_LOOP_TAGS, dimension_indices)

    transfer_loop.add_tag(transfer_loop_tags)

    for dim_index in range(1, len(dimensions)):
        dim_1 = dim_index
        dim_2 = dim_index + 1

        transfer_data = {
            DIMENSION__DIMENSION_INDEX.format(**{DIMENSION_INDEX: dim_1}): dim_1,
            DIMENSION__DIMENSION_INDEX.format(**{DIMENSION_INDEX: dim_2}): dim_2,
            TRANSFER_TYPE: ONE_BOND,
            IS_INDIRECT: NEF_FALSE,
        }

        transfer_loop.add_data(
            [
                transfer_data,
            ]
        )

    peak_loop = Loop.from_scratch(SPECTRUM_PEAK_LOOP_CATEGORY)
    frame.add_loop(peak_loop)

    peak_loop_tags = _expand_templates(PEAK_LOOP_TAGS, dimension_indices)

    if have_comments:
        peak_loop_tags.append(CCPN_COMMENT)

    if have_merits:
        peak_loop_tags.append(CCPN_MERIT)

    peak_loop.add_tag(peak_loop_tags)

    for index, peak in enumerate(peaks):
        peak_data = {
            INDEX: index,
            PEAK_ID: index if peak.id is None else peak.id,
            HEIGHT: peak.height,
            HEIGHT_UNCERTAINTY: peak.height_uncertainty,
            VOLUME: peak.volume,
            VOLUME_UNCERTAINTY: peak.volume_uncertainty,
        }

        for dim_index, shift in enumerate(peak.shifts, start=1):
            peak_data[CHAIN_CODE__DIMENSION_INDEX.format(dimension_index=dim_index)] = (
                shift.atom.residue.chain_code
            )
            peak_data[
                SEQUENCE_CODE__DIMENSION_INDEX.format(dimension_index=dim_index)
            ] = shift.atom.residue.sequence_code
            peak_data[
                RESIDUE_NAME__DIMENSION_INDEX.format(dimension_index=dim_index)
            ] = shift.atom.residue.residue_name
            peak_data[ATOM_NAME__DIMENSION_INDEX.format(dimension_index=dim_index)] = (
                shift.atom.atom_name
            )

            peak_data[POSITION__DIMENSION_INDEX.format(dimension_index=dim_index)] = (
                shift.value
            )
            peak_data[
                POSITION_UNCERTAINTY__DIMENSION_INDEX.format(dimension_index=dim_index)
            ] = shift.value_uncertainty

            if have_comments:
                peak_data[CCPN_COMMENT] = (
                    peak.comment if peak.comment is not None else UNUSED
                )
            if have_merits:
                peak_data[CCPN_MERIT] = (
                    peak.figure_of_merit if peak.figure_of_merit is not None else UNUSED
                )
        peak_loop.add_data(
            [
                peak_data,
            ]
        )

    return frame


# TODO: test empt frames and loops
def frame_to_peaks(frame: Saveframe, source="unknown") -> List[NewPeak]:
    """
    convert a NEF spectrum frame to a list of peaks

    :param frame: the spectrum save frame
    :param source: the source of the save frame (for error reporting)
    :return: a list of Peak structures
    """
    loop = frame.get_loop("nef_peak")

    num_dimensions = _get_peak_loop_dimensions(loop)

    peaks = []

    for line_number, row in enumerate(loop_row_dict_iter(loop), start=1):
        shift_data = []
        for dim_index in range(1, num_dimensions + 1):

            values = {}
            fields = [CHAIN_CODE, SEQUENCE_CODE, RESIDUE_NAME, ATOM_NAME]
            raw_tags = [
                CHAIN_CODE__DIMENSION_INDEX,
                SEQUENCE_CODE__DIMENSION_INDEX,
                RESIDUE_NAME__DIMENSION_INDEX,
                ATOM_NAME__DIMENSION_INDEX,
            ]
            dimension_tags = [
                raw_tag.format(dimension_index=dim_index) for raw_tag in raw_tags
            ]
            for name, tag in zip(fields, dimension_tags):
                value = unused_to_empty_string(row[tag])
                values[name] = value

            residue_values = {
                key: value
                for key, value in values.items()
                if not key.startswith(ATOM_NAME)
            }

            residue = Residue(**residue_values)

            if values[ATOM_NAME] == "":
                values[ATOM_NAME] = "?"

            atom_label = AtomLabel(residue, values[ATOM_NAME])

            position = row[POSITION__DIMENSION_INDEX.format(dimension_index=dim_index)]

            line_info = LineInfo(
                f"{source}[{frame.name} ]", line_number, _row_to_table(row)
            )
            _raise_if_position_isnt_float(position, line_info)

            position_uncertainty = unused_to_none(
                row[
                    POSITION_UNCERTAINTY__DIMENSION_INDEX.format(
                        dimension_index=dim_index
                    )
                ]
            )

            shift_datum = ShiftData(atom_label, position, position_uncertainty)

            shift_data.append(shift_datum)

        height_volume_data = {}
        for name in [VOLUME, VOLUME_UNCERTAINTY, HEIGHT, HEIGHT_UNCERTAINTY]:

            value = unused_to_none(row[name])
            height_volume_data[name] = unused_to_none(value)

        peak = NewPeak(tuple(shift_data), **height_volume_data)

        peaks.append(peak)

    return peaks


def _raise_if_position_isnt_float(value: str, line_info: LineInfo):
    """
    raise an exception if a value which can't be converted to a float
    :param value: the value to check for conversion to a float
    :param line_info: where the error occured
    """
    if not is_float(value):
        msg = f"""
                    in the spectrum save frame  in {line_info.file_name} at row {line_info.line_number}
                    the position wasn't a float: {value}

                    processed values from row are:

                """

        msg = dedent(msg)

        msg += line_info.line

        raise BadPositionException(msg)


def _get_peak_loop_dimensions(loop: Loop) -> int:
    """
    from the peak loop of a spectrum work out the number of dimensions by counting the number of position_xxx tags

    Note: we do it this way rather than depending on the _nef_spectrum_dimension's as this is the more fundamental data
    :param loop: the peak loop
    :return: the number of dimensions detected
    """

    num_dimensions = 0
    for tag in loop.tags:
        tag_fields = tag.split("_")
        if (
            len(tag_fields) == 2
            and tag_fields[0] == "position"
            and is_int(tag_fields[1])
        ):
            num_dimensions += 1
    return num_dimensions



import typer

from nef_pipelines.lib.structures import AtomLabel, Peak, PeakValues, SequenceResidue
from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.nmrview.exporters.peaks import (
    _make_names_unique,
    _row_to_peak,
    peaks,
)

app = typer.Typer()
app.command()(peaks)

EXPECTED = open(
    path_in_test_data(__file__, "expected_sequence_and_peaks_nmrview.txt")
).read()

EXPECTED_AXES = ["1H_1", "1H_2", "15N"]


def test_make_names_unique():
    TEST = ["1H", "1H", "15N"]

    result = _make_names_unique(TEST)

    assert result == EXPECTED_AXES


def test_row_to_peak():
    TEST_ROW = {
        "index": 1,
        "peak_id": 0,
        "volume": 0.38,
        "volume_uncertainty": ".",
        "height": 0.38,
        "height_uncertainty": ".",
        "position_1": 10.405,
        "position_uncertainty_1": ".",
        "position_2": 8.796,
        "position_uncertainty_2": ".",
        "position_3": 132.49,
        "position_uncertainty_3": ".",
        "chain_code_1": "A",
        "sequence_code_1": 1,
        "residue_name_1": "ala",
        "atom_name_1": "HE1",
        "chain_code_2": "A",
        "sequence_code_2": 3,
        "residue_name_2": "ala",
        "atom_name_2": "HN",
        "chain_code_3": "A",
        "sequence_code_3": 1,
        "residue_name_3": "ala",
        "atom_name_3": "N",
    }

    result = _row_to_peak(EXPECTED_AXES, TEST_ROW)

    assignments = {
        "15N": [
            AtomLabel(
                SequenceResidue(chain_code="A", sequence_code=1, residue_name="ala"),
                atom_name="N",
            )
        ],
        "1H_1": [
            AtomLabel(
                SequenceResidue(chain_code="A", sequence_code=1, residue_name="ala"),
                atom_name="HE1",
            )
        ],
        "1H_2": [
            AtomLabel(
                SequenceResidue(chain_code="A", sequence_code=3, residue_name="ala"),
                atom_name="HN",
            )
        ],
    }
    positions = {"1H_1": 10.405, "1H_2": 8.796, "15N": 132.49}

    EXPECTED_PEAK = Peak(
        id=0,
        positions=positions,
        assignments=assignments,
        values=PeakValues(serial=1, volume=0.38, height=0.38),
    )

    assert result == EXPECTED_PEAK


EXPECTED = open(path_in_test_data(__file__, "expected_nmr_view.xpk")).read()


# noinspection PyUnusedLocal
def test_3peaks(clear_cache):

    STREAM = open(path_in_test_data(__file__, "nef_3_peaks.nef")).read()

    result = run_and_report(app, [], input=STREAM)

    assert_lines_match(EXPECTED, result.stdout)



import typer
from pynmrstar import Entry

from nef_pipelines.lib.test_lib import path_in_test_data, run_and_report
from nef_pipelines.tools.frames.rename import rename


app = typer.Typer()
app.command()(filter)

def test_filter_unassigned():
    ...



from itertools import chain, cycle, islice, zip_longest
from pathlib import Path
from typing import Iterable, List

import typer
from fastaparser import Reader
from ordered_set import OrderedSet

from nef_pipelines.lib.sequence_lib import (
    BadResidue,
    MoleculeType,
    MoleculeTypes,
    chain_code_iter,
    offset_chain_residues,
    sequence_3let_to_res,
    sequence_to_nef_frame,
    translate_1_to_3,
)
from nef_pipelines.lib.structures import SequenceResidue
from nef_pipelines.lib.typer_utils import get_args
from nef_pipelines.lib.util import (
    STDIN,
    exit_error,
    parse_comma_separated_options,
    process_stream_and_add_frames,
)
from nef_pipelines.transcoders.fasta import import_app

app = typer.Typer()

NO_CHAIN_START_HELP = """don't include the start chain link type on a chain for the first residue [linkage will be
                         middle] for the named chains. Either use a comma joined list of chains [e.g. A,B] or call this
                         option multiple times to set chain starts for multiple chains"""
NO_CHAIN_END_HELP = """don't include the end chain link type on a chain for the last residue [linkage will be
                       middle] for the named chains. Either use a comma joined list of chains [e.g. A,B] or call this
                       option multiple times to set chain ends for multiple chains"""


# todo add comment to other sequences etc
@import_app.command()
def sequence(
    chain_codes: List[str] = typer.Option(
        [],
        "--chains",
        help="chain codes to use for the imported chains, can be a a comma separated list or can be called "
        "multiple times",
        metavar="<CHAIN-CODES>",
    ),
    starts: List[str] = typer.Option(
        [],
        "--starts",
        help="first residue number of sequences can be a comma separated list or ",
        metavar="<START>",
    ),
    no_chain_starts: List[str] = typer.Option(
        [], "--no-chain-start", help=NO_CHAIN_START_HELP
    ),
    no_chain_ends: List[str] = typer.Option(
        [], "--no-chain-end", help=NO_CHAIN_END_HELP
    ),
    # TODO: unused inputs!
    entry_name: str = typer.Option("fasta", help="a name for the entry if required"),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--in",
        help="pipe to read NEF data from, for testing [overrides stdin !use stdin instead!]",
    ),
    molecule_types: List[MoleculeType] = typer.Option(
        [
            MoleculeType.PROTEIN,
        ],
        "--molecule-type",
        help="molecule type",
    ),
    file_names: List[Path] = typer.Argument(
        ..., help="the file to read", metavar="<FASTA-FILE>"
    ),
):
    """- convert fasta sequence to nef"""

    chain_codes = parse_comma_separated_options(chain_codes)
    if not chain_codes:
        chain_codes = ["A"]

    no_chain_starts = parse_comma_separated_options(no_chain_starts)
    if not no_chain_starts:
        no_chain_starts = [False]

    no_chain_ends = parse_comma_separated_options(no_chain_ends)
    if no_chain_ends:
        no_chain_ends = [False]

    starts = [int(elem) for elem in parse_comma_separated_options(starts)]
    if not starts:
        starts = [1]

    molecule_types = parse_comma_separated_options(molecule_types)
    if not molecule_types:
        molecule_types = [
            MoleculeType.PROTEIN,
        ]

    args = get_args()

    args.chain_codes = chain_codes
    args.no_chain_starts = no_chain_starts
    args.no_chain_ends = no_chain_ends
    args.starts = starts
    args.molecule_types = molecule_types

    # TODO: doesn't provide a pipe command
    process_sequences(args)


def process_sequences(args):
    fasta_frames = []

    chain_code_iterator = chain_code_iter(args.chain_codes)
    fasta_sequences = OrderedSet()
    for file_path in args.file_names:
        fasta_sequences.update(
            read_sequences(file_path, chain_code_iterator, args.molecule_types)
        )

    read_chain_codes = residues_to_chain_codes(fasta_sequences)

    offsets = _get_sequence_offsets(read_chain_codes, args.starts)

    fasta_sequences = offset_chain_residues(fasta_sequences, offsets)

    fasta_sequences = sorted(fasta_sequences)

    fasta_frames.append(sequence_to_nef_frame(fasta_sequences))

    entry = process_stream_and_add_frames(fasta_frames, args)

    print(entry)


def _get_sequence_offsets(chain_codes: List[str], starts: List[int]):

    offsets = [start - 1 for start in starts]
    cycle_starts = chain(
        offsets,
        cycle(
            [
                0,
            ]
        ),
    )
    offsets = list(islice(cycle_starts, len(chain_codes)))

    return {chain_code: offset for chain_code, offset in zip(chain_codes, offsets)}


def residues_to_chain_codes(residues: List[SequenceResidue]) -> List[str]:
    return list(OrderedSet([residue.chain_code for residue in residues]))


# could do with taking a list of offsets
# noinspection PyUnusedLocal
def read_sequences(
    path: Path, chain_codes: Iterable[str], molecule_types: List[MoleculeType]
) -> List[SequenceResidue]:

    residues = OrderedSet()
    try:
        with open(path) as handle:
            try:
                reader = Reader(handle)
                sequences = [sequence for sequence in reader]
            except Exception as e:
                # check if relative to os.getcwd
                exit_error(f"Error reading fasta file {str(path)}", e)

            number_sequences = len(sequences)

            # read as many chain codes as there are sequences
            # https://stackoverflow.com/questions/16188270/get-a-fixed-number-of-items-from-a-generator

            chain_codes = list(islice(chain_codes, number_sequences))

            for sequence, chain_code, molecule_type in zip_longest(
                sequences, chain_codes, molecule_types, fillvalue=None
            ):
                if molecule_type is None:
                    molecule_type = MoleculeTypes.PROTEIN

                sequence = [letter_code.letter_code for letter_code in sequence]
                try:
                    sequence_3_let = translate_1_to_3(
                        sequence, molecule_type=molecule_type
                    )
                except BadResidue as e:
                    exit_error(
                        f"Error translating sequence {sequence} to 3 letter code {e}",
                        e,
                    )
                chain_residues = sequence_3let_to_res(sequence_3_let, chain_code)

                residues.update(chain_residues)

    except IOError as e:
        exit_error(f"couldn't open {path} because:\n{e}", e)

    return residues



import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.mars.exporters.shifts import shifts

app = typer.Typer()
app.command()(shifts)


def test_export_shifts_nef5_short():
    stream = open(path_in_test_data(__file__, "sec5_short.neff")).read()

    result = run_and_report(app, ["-"], input=stream)

    EXPECTED = """\
              H      N        CA      CA-1    CB      CB-1
        PR_1  8.594  133.252  61.159  59.291  33.274  37.309
        PR_2  8.928  131.397  -       -       -       -
        PR_3  9.671  130.698  60.735  62.418  39.792  69.744
        PR_4  8.185  128.894  53.744  54.679  43.207  34.050
        PR_5  8.548  128.09   56.292  65.574  32.964  69.899
    """

    assert_lines_match(EXPECTED, result.stdout)



import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.talos.exporters.shifts import shifts

app = typer.Typer()
app.command()(shifts)

EXPECTED = """
REMARK Chemical shift table for nmrpipe

DATA CHAIN A
DATA FIRST_RESID 1

DATA SEQUENCE MQIF

VARS   RESID    RESNAME    ATOMNAME      SHIFT
FORMAT %4d      %1s        %4s           %8.3f
1               M          CB           33.27
1               M          CA           54.45
1               M          C           170.54
1               M          HA            4.23
2               Q          CA           55.08
2               Q          CB           30.76
2               Q          C           175.92
2               Q          HA            5.249
2               Q          N           123.22
3               I          HA            4.213
3               I          C           172.45
3               I          CA           59.57
3               I          N           115.34
3               I          CB           42.21
4               F          HA            5.63
4               F          C           175.32
4               F          CB           41.48
4               F          CA           55.21
4               F          N           118.11
"""


def test_4peaks(clear_cache):

    STREAM = open(path_in_test_data(__file__, "ubi_4.nef")).read()

    result = run_and_report(app, [], input=STREAM)

    assert_lines_match(EXPECTED, result.stdout)


# TODO check filtering & bad inputs

EXPECTED_PROTONATED_HIS = """
REMARK Chemical shift table for nmrpipe

DATA CHAIN A
DATA FIRST_RESID 1

DATA SEQUENCE MhIF

VARS   RESID    RESNAME    ATOMNAME      SHIFT
FORMAT %4d      %1s        %4s           %8.3f
1               M          CB           33.27
1               M          CA           54.45
1               M          C           170.54
1               M          HA            4.23
2               h          CA           55.08
2               h          CB           30.76
2               h          C           175.92
2               h          HA            5.249
2               h          N           123.22
3               I          HA            4.213
3               I          C           172.45
3               I          CA           59.57
3               I          N           115.34
3               I          CB           42.21
4               F          HA            5.63
4               F          C           175.32
4               F          CB           41.48
4               F          CA           55.21
4               F          N           118.11
"""


def test_4peak_his_he(clear_cache):

    STREAM = open(path_in_test_data(__file__, "protonated_his.nef")).read()

    result = run_and_report(app, [], input=STREAM)

    assert_lines_match(EXPECTED_PROTONATED_HIS, result.stdout)


EXPECTED_OXIDISED_CYS = """
REMARK Chemical shift table for nmrpipe

DATA CHAIN A
DATA FIRST_RESID 1

DATA SEQUENCE McIF

VARS   RESID    RESNAME    ATOMNAME      SHIFT
FORMAT %4d      %1s        %4s           %8.3f
1               M          CB           33.27
1               M          CA           54.45
1               M          C           170.54
1               M          HA            4.23
2               c          CA           55.08
2               c          CB           30.76
2               c          C           175.92
2               c          HA            5.249
2               c          N           123.22
3               I          HA            4.213
3               I          C           172.45
3               I          CA           59.57
3               I          N           115.34
3               I          CB           42.21
4               F          HA            5.63
4               F          C           175.32
4               F          CB           41.48
4               F          CA           55.21
4               F          N           118.11
"""


def test_4peak_cis_oxidised(clear_cache):

    STREAM = open(path_in_test_data(__file__, "oxidised_cys.nef")).read()

    result = run_and_report(app, [], input=STREAM)

    assert_lines_match(EXPECTED_OXIDISED_CYS, result.stdout)


EXPECTED_OFFSET_1 = """
    REMARK Chemical shift table for nmrpipe

    DATA CHAIN A
    DATA FIRST_RESID 2

    DATA SEQUENCE MQIF

    VARS   RESID    RESNAME    ATOMNAME      SHIFT
    FORMAT %4d      %1s        %4s           %8.3f
    2               M          CB           33.27
    2               M          CA           54.45
    2               M          C           170.54
    2               M          HA            4.23
    3               Q          CA           55.08
    3               Q          CB           30.76
    3               Q          C           175.92
    3               Q          HA            5.249
    3               Q          N           123.22
    4               I          HA            4.213
    4               I          C           172.45
    4               I          CA           59.57
    4               I          N           115.34
    4               I          CB           42.21
    5               F          HA            5.63
    5               F          C           175.32
    5               F          CB           41.48
    5               F          CA           55.21
    5               F          N           118.11
"""


def test_4peaks_offset(clear_cache):
    STREAM = open(path_in_test_data(__file__, "ubi_4_offset_1.nef")).read()

    result = run_and_report(app, [], input=STREAM)

    assert_lines_match(EXPECTED_OFFSET_1, result.stdout)



import typer

from nef_pipelines.lib.test_lib import (
    NOQA_E501,
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.xplor.importers.dihedrals import dihedrals

app = typer.Typer()
app.command()(dihedrals)


# noinspection PyUnusedLocal
def test_2_dihedrals():
    sequence_path = path_in_test_data(__file__, "3a_ab.neff")
    dihedrals_path = path_in_test_data(__file__, "test_2_dihedrals.tbl")

    nef_sequence = open(sequence_path, "r").read()

    args = [dihedrals_path]
    result = run_and_report(app, args, input=nef_sequence)

    EXPECTED = """\
        save_nef_dihedral_restraint_list_test_2_dihedrals                               # noqa: E501
            _nef_dihedral_restraint_list.sf_category     nef_dihedral_restraint_list
            _nef_dihedral_restraint_list.sf_framecode    nef_dihedral_restraint_list_test_2_dihedrals
            _nef_dihedral_restraint_list.potential_type  undefined

            loop_
                _nef_dihedral_restraint.index
                _nef_dihedral_restraint.restraint_id
                _nef_dihedral_restraint.restraint_combination_id
                _nef_dihedral_restraint.chain_code_1
                _nef_dihedral_restraint.sequence_code_1
                _nef_dihedral_restraint.residue_name_1
                _nef_dihedral_restraint.atom_name_1
                _nef_dihedral_restraint.chain_code_2
                _nef_dihedral_restraint.sequence_code_2
                _nef_dihedral_restraint.residue_name_2
                _nef_dihedral_restraint.atom_name_2
                _nef_dihedral_restraint.chain_code_3
                _nef_dihedral_restraint.sequence_code_3
                _nef_dihedral_restraint.residue_name_3
                _nef_dihedral_restraint.atom_name_3
                _nef_dihedral_restraint.chain_code_4
                _nef_dihedral_restraint.sequence_code_4
                _nef_dihedral_restraint.residue_name_4
                _nef_dihedral_restraint.atom_name_4
                _nef_dihedral_restraint.weight
                _nef_dihedral_restraint.target_value
                _nef_dihedral_restraint.lower_limit
                _nef_dihedral_restraint.upper_limit

                1   1   .   AAAA   1   .   C   AAAA   2   .   N    AAAA   2   .   CA   AAAA   2   .   C   1.0   -45.7   -166.2   74.8
                2   2   .   AAAA   2   .   N   AAAA   2   .   CA   AAAA   2   .   C    AAAA   3   .   N   1.0   65.4    -55.3    186.1

            stop_
        save_
    """.replace(
        NOQA_E501, ""
    )

    result = isolate_frame(
        result.stdout, "nef_dihedral_restraint_list_test_2_dihedrals"
    )

    assert_lines_match(EXPECTED, result)


def test_2_dihedrals_bad():
    sequence_path = path_in_test_data(__file__, "3a_ab.neff")
    dihedrals_path = path_in_test_data(__file__, "test_2_dihedrals_bad.tbl")

    nef_sequence = open(sequence_path, "r").read()

    args = [dihedrals_path]
    result = run_and_report(app, args, input=nef_sequence, expected_exit_code=1)

    assert "ERROR" in result.stdout
    assert "failed to read dihedral restraints" in result.stdout


def test_2_dihedrals_no_segids():
    sequence_path = path_in_test_data(__file__, "3a_ab.neff")
    dihedrals_path = path_in_test_data(__file__, "test_2_dihedrals_no_segids.tbl")

    nef_sequence = open(sequence_path, "r").read()

    args = [dihedrals_path, "--chains", "AAAA"]
    result = run_and_report(app, args, input=nef_sequence)

    EXPECTED = """\
        save_nef_dihedral_restraint_list_test_2_dihedrals_no_segids                              # noqa: E501
            _nef_dihedral_restraint_list.sf_category     nef_dihedral_restraint_list
            _nef_dihedral_restraint_list.sf_framecode    nef_dihedral_restraint_list_test_2_dihedrals_no_segids
            _nef_dihedral_restraint_list.potential_type  undefined

            loop_
                _nef_dihedral_restraint.index
                _nef_dihedral_restraint.restraint_id
                _nef_dihedral_restraint.restraint_combination_id
                _nef_dihedral_restraint.chain_code_1
                _nef_dihedral_restraint.sequence_code_1
                _nef_dihedral_restraint.residue_name_1
                _nef_dihedral_restraint.atom_name_1
                _nef_dihedral_restraint.chain_code_2
                _nef_dihedral_restraint.sequence_code_2
                _nef_dihedral_restraint.residue_name_2
                _nef_dihedral_restraint.atom_name_2
                _nef_dihedral_restraint.chain_code_3
                _nef_dihedral_restraint.sequence_code_3
                _nef_dihedral_restraint.residue_name_3
                _nef_dihedral_restraint.atom_name_3
                _nef_dihedral_restraint.chain_code_4
                _nef_dihedral_restraint.sequence_code_4
                _nef_dihedral_restraint.residue_name_4
                _nef_dihedral_restraint.atom_name_4
                _nef_dihedral_restraint.weight
                _nef_dihedral_restraint.target_value
                _nef_dihedral_restraint.lower_limit
                _nef_dihedral_restraint.upper_limit

                1   1   .   AAAA   1   .   C   AAAA   2   .   N    AAAA   2   .   CA   AAAA   2   .   C   1.0   -45.7   -166.2   74.8
                2   2   .   AAAA   2   .   N   AAAA   2   .   CA   AAAA   2   .   C    AAAA   3   .   N   1.0   65.4    -55.3    186.1

            stop_
        save_
    """.replace(
        NOQA_E501, ""
    )

    result = isolate_frame(
        result.stdout, "nef_dihedral_restraint_list_test_2_dihedrals_no_segids"
    )

    assert_lines_match(EXPECTED, result)


def test_2_dihedrals_overriding_segids():
    sequence_path = path_in_test_data(__file__, "3a_ab.neff")
    dihedrals_path = path_in_test_data(__file__, "test_2_dihedrals.tbl")

    nef_sequence = open(sequence_path, "r").read()

    args = [dihedrals_path, "--chains", "BBBB", "--use-chains"]
    result = run_and_report(app, args, input=nef_sequence)

    EXPECTED = """\
        save_nef_dihedral_restraint_list_test_2_dihedrals                                     # noqa: E501
            _nef_dihedral_restraint_list.sf_category     nef_dihedral_restraint_list
            _nef_dihedral_restraint_list.sf_framecode    nef_dihedral_restraint_list_test_2_dihedrals
            _nef_dihedral_restraint_list.potential_type  undefined

            loop_
                _nef_dihedral_restraint.index
                _nef_dihedral_restraint.restraint_id
                _nef_dihedral_restraint.restraint_combination_id
                _nef_dihedral_restraint.chain_code_1
                _nef_dihedral_restraint.sequence_code_1
                _nef_dihedral_restraint.residue_name_1
                _nef_dihedral_restraint.atom_name_1
                _nef_dihedral_restraint.chain_code_2
                _nef_dihedral_restraint.sequence_code_2
                _nef_dihedral_restraint.residue_name_2
                _nef_dihedral_restraint.atom_name_2
                _nef_dihedral_restraint.chain_code_3
                _nef_dihedral_restraint.sequence_code_3
                _nef_dihedral_restraint.residue_name_3
                _nef_dihedral_restraint.atom_name_3
                _nef_dihedral_restraint.chain_code_4
                _nef_dihedral_restraint.sequence_code_4
                _nef_dihedral_restraint.residue_name_4
                _nef_dihedral_restraint.atom_name_4
                _nef_dihedral_restraint.weight
                _nef_dihedral_restraint.target_value
                _nef_dihedral_restraint.lower_limit
                _nef_dihedral_restraint.upper_limit

                1   1   .   BBBB   1   .   C   BBBB   2   .   N    BBBB   2   .   CA   BBBB   2   .   C   1.0   -45.7   -166.2   74.8
                2   2   .   BBBB   2   .   N   BBBB   2   .   CA   BBBB   2   .   C    BBBB   3   .   N   1.0   65.4    -55.3    186.1

            stop_
        save_
    """.replace(
        NOQA_E501, ""
    )

    result = isolate_frame(
        result.stdout, "nef_dihedral_restraint_list_test_2_dihedrals"
    )

    assert_lines_match(EXPECTED, result)



import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.pales.exporters.template import template

app = typer.Typer()
app.command()(template)

SEQUENCE_STREAM = open(path_in_test_data(__file__, "tailin_seq_short.nef")).read()


# noinspection PyUnusedLocal
def test_template(clear_cache):

    result = run_and_report(app, [], input=SEQUENCE_STREAM)

    if result.exit_code != 0:
        print("INFO: stdout from failed read:\n", result.stdout)

    assert result.exit_code == 0

    # note: sequence starts at residue 10 so residue 10 has an amine not an amide and isn't in the tremplate...
    EXPECTED = """\

        VARS    RESID_I  RESNAME_I  ATOMNAME_I  RESID_J  RESNAME_J  ATOMNAME_J  D      DD     W
        FORMAT  %5d      %6s        %6s         %5d      %6s        %6s         %9.3f  %9.3f  %.2f
                11       TYR        HN          11       TYR        N           0.0    0.0    1.0
                12       ALA        HN          12       ALA        N           0.0    0.0    1.0
                13       GLN        HN          13       GLN        N           0.0    0.0    1.0
                15       ARG        HN          15       ARG        N           0.0    0.0    1.0
                16       LEU        HN          16       LEU        N           0.0    0.0    1.0
                17       ARG        HN          17       ARG        N           0.0    0.0    1.0
                18       LEU        HN          18       LEU        N           0.0    0.0    1.0
                19       GLY        HN          19       GLY        N           0.0    0.0    1.0
                20       PHE        HN          20       PHE        N           0.0    0.0    1.0
                21       GLU        HN          21       GLU        N           0.0    0.0    1.0
                22       ASP        HN          22       ASP        N           0.0    0.0    1.0
    """

    assert_lines_match(EXPECTED, result.stdout)



import typer
from pynmrstar import Entry

from nef_pipelines import nef_app
from nef_pipelines.lib.constants import NEF_PIPELINES
from nef_pipelines.lib.header_lib import create_header_frame
from nef_pipelines.lib.typer_utils import get_args
from nef_pipelines.lib.util import get_version, script_name


# noinspection PyUnusedLocal
@nef_app.app.command()
def header(
    name: str = typer.Argument("nef", help="name for the entry", metavar="<ENTRY-NAME>")
):
    """-  add a header to the stream"""
    args = get_args()

    entry = build_meta_data(args)

    print(entry)


def build_meta_data(args):

    version = get_version()
    result = Entry.from_scratch(args.name)
    header_frame = create_header_frame(NEF_PIPELINES, version, script_name(__file__))
    result.add_saveframe(header_frame)

    return result


if __name__ == "__main__":
    typer.run(header)



import sys
from dataclasses import dataclass, replace
from enum import Enum
from pathlib import Path
from textwrap import dedent
from typing import List, Union

import hjson
import typer
from fyeah import f
from ordered_set import OrderedSet
from pynmrstar import Entry
from tabulate import tabulate

from nef_pipelines.lib.nef_lib import (
    UNUSED,
    add_frames_to_entry,
    loop_row_namespace_iter,
    read_entry_from_file_or_exit_error,
    read_entry_from_file_or_stdin_or_exit_error,
)
from nef_pipelines.lib.sequence_lib import chain_code_iter, sequence_from_entry_or_exit
from nef_pipelines.lib.shift_lib import shifts_to_nef_frame
from nef_pipelines.lib.structures import AtomLabel, Residue, ShiftData, ShiftList
from nef_pipelines.lib.translation.chem_comp import ID, ChemCompLinking
from nef_pipelines.lib.util import (
    STDIN,
    chunks,
    exit_error,
    flatten,
    nef_pipelines_root,
    parse_comma_separated_options,
)
from nef_pipelines.transcoders.nmrstar import import_app


class StereoAssignmentHandling(Enum):
    ALL_AMBIGUOUS = "all-ambiguous"
    AS_ASSIGNED = "as-assigned"
    AUTO = "auto"

    def __str__(self):
        return self._name_.lower().replace("_", "-")


STEREO_HELP = """\
    how to handle stereo assignments the choices are:
    - ambiguous: assume all stereo assignments are ambiguous
    - as-assigned: use the stereo assignments as they are in the file
    - auto: use as assigned if some geminal stereo assignments are present, otherwise assume all are ambiguous
"""


@import_app.command()
def shifts(
    chain_codes: List[str] = typer.Option(
        None,
        "--chains",
        help="chain codes as a list of names separated by commas, repeated calls will add further chains [default A]",
        metavar="<CHAIN-CODES>",
    ),
    frame_name: str = typer.Option(
        None,
        "-f",
        "--frame-name",
        help="a name for the frame, default is the name nmrstar entry",
    ),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--input",
        metavar="|PIPE|",
        help="file to read NEF data from [- is stdin; defaults is stdin]]",
    ),
    stereo_mode: StereoAssignmentHandling = typer.Option("auto", help=STEREO_HELP),
    use_author: bool = typer.Option(
        False, help="use author fields for chain_code, sequence_code and residue_name"
    ),
    file_path: Path = typer.Argument(
        ..., help="input files of type shifts.txt", metavar="<NMR-STAR-shifts>.str"
    ),
):
    """- convert nmrpipe shift file <nmr-star>.str to NEF [alpha]"""

    nmrstar_entry = read_entry_from_file_or_exit_error(file_path)

    chain_codes = parse_comma_separated_options(chain_codes)

    nef_entry = read_entry_from_file_or_stdin_or_exit_error(input)

    nef_entry = pipe(
        nef_entry,
        chain_codes,
        frame_name,
        nmrstar_entry,
        file_path,
        use_author,
        stereo_mode,
    )

    print(nef_entry)


def _get_chem_atom_set_atoms(chem_atom_set, chem_comp):
    chem_atom_set_by_id = {
        chem_atom_set.ID: chem_atom_set for chem_atom_set in chem_comp.chemAtomSets
    }

    if isinstance(chem_atom_set, ID):
        result = [chem_atom_set]
    else:
        result = list(chem_atom_set.chemAtoms) if chem_atom_set.chemAtoms else []

        if chem_atom_set.chemAtomSets:
            for nested_set in chem_atom_set.chemAtomSets:
                result.extend(chem_atom_set_by_id[nested_set].chemAtoms)

    return set(result)


def _get_geminal_pairs(residue_names):

    ambiguity_translations_path = (
        Path(nef_pipelines_root())
        / "nef_pipelines"
        / "data"
        / "ambiguity_translations.json"
    )

    with open(ambiguity_translations_path, "r") as f:
        all_prochiral_pairs = hjson.loads(f.read())

    # import nef_pipelines.lib.translation.io as converter_io
    # converter_io.load_chem_comps()
    # all_prochiral_pairs = {}
    #
    # # for residue_name in residue_names:
    # residue_names = [chem_comp for chem_comp in converter_io.CHEM_COMPS]
    for residue_name in residue_names:
        if residue_name in all_prochiral_pairs:
            continue

        # TODO:  this would be a lot neater if data was only loaded on first access
        import nef_pipelines.lib.translation.io as converter_io

        converter_io.load_chem_comps()

        # TODO: add the ability to add unknown chem comps
        _exit_if_unknown_chem_comp(residue_name)

        residue_linkings = {}
        all_prochiral_pairs[residue_name] = residue_linkings

        converter_io.load_chem_comps()

        _exit_if_unknown_chem_comp(residue_name)

        chem_comp = converter_io.CHEM_COMPS[residue_name]

        chem_atoms_by_id = {
            chem_comp_atom.ID: chem_comp_atom for chem_comp_atom in chem_comp.chemAtoms
        }
        atom_sets_by_id = {
            chem_atom_set.ID: chem_atom_set for chem_atom_set in chem_comp.chemAtomSets
        }

        active_chem_comp_vars = [
            chem_comp_var
            for chem_comp_var in chem_comp.chemCompVars
            if chem_comp_var.isDefaultVar
            and chem_comp_var.linking in ["start", "middle", "end"]
        ]

        chem_atom_set_atoms = {}
        atom_sets_by_key = {
            (chem_atom_set.name, chem_atom_set.subType): chem_atom_set
            for chem_atom_set in chem_comp.chemAtomSets
        }
        for chem_atom_set in chem_comp.chemAtomSets:
            chem_atom_set_atoms[(chem_atom_set.name, chem_atom_set.subType)] = (
                _get_chem_atom_set_atoms(chem_atom_set, chem_comp)
            )

        active_chem_atom_sets_by_linking = {}
        for chem_comp_var in active_chem_comp_vars:
            linking = chem_comp_var.linking
            active_chem_atom_sets_by_linking[chem_comp_var.linking.value] = []
            active_chem_comp_atom_ids = set(chem_comp_var.chemAtoms)
            for atom_set_name, atom_set_atom_ids in chem_atom_set_atoms.items():
                if atom_set_atom_ids.issubset(active_chem_comp_atom_ids):
                    active_chem_atom_sets_by_linking[linking].append(atom_set_name)

        for linking, active_atom_set_keys in active_chem_atom_sets_by_linking.items():
            new_and_old_names = all_prochiral_pairs[residue_name].setdefault(
                linking, {}
            )
            for atom_set_key in active_atom_set_keys:
                chem_atom_set = atom_sets_by_key[atom_set_key]
                if chem_atom_set.isEquivalent:
                    continue

                chem_atoms = (
                    [chem_atoms_by_id[id].name for id in chem_atom_set.chemAtoms]
                    if chem_atom_set.chemAtoms
                    else []
                )
                atom_sets = (
                    [atom_sets_by_id[id].name for id in chem_atom_set.chemAtomSets]
                    if chem_atom_set.chemAtomSets
                    else []
                )

                if "|" in chem_atom_set.name:
                    continue

                active_pair = sorted(chem_atoms or atom_sets)

                replacements = ["x", "y"]

                # TODO: will need to patch for ''s etc
                pair_values = new_and_old_names.setdefault(chem_atom_set.name, {})
                for i, name in enumerate(active_pair):
                    if name[-1] == "*":
                        new_name = f"{name[:-2]}{replacements[i]}%"
                    else:
                        new_name = f"{name[:-1]}{replacements[i]}"
                    pair_values[name] = new_name

        # print(residue_name, all_prochiral_pairs[residue_name])
    # import json
    # with open(ambiguity_translations_path, 'w') as fh:
    #     fh.write(json.dumps(all_prochiral_pairs))
    # print(f'wrote {ambiguity_translations_path}')
    # sys.exit(0)
    # print(all_prochiral_pairs.keys())
    return all_prochiral_pairs


def _get_atom_sets_by_residue(residue_names):

    equivalent_atoms_path = (
        Path(nef_pipelines_root())
        / "nef_pipelines"
        / "data"
        / "default_atom_sets_by_comp_and_linking.json"
    )
    with open(equivalent_atoms_path, "r") as f:
        equivalent_atoms_by_residue = hjson.loads(f.read())

    for residue_name in residue_names:

        if residue_name in equivalent_atoms_by_residue:
            continue

        # TODO:  this would be a lot neater if data was only loaded on first access
        import nef_pipelines.lib.translation.io as converter_io

        converter_io.load_chem_comps()

        # TODO: add the ability to add unknown chem comps
        _exit_if_unknown_chem_comp(residue_name)

        chem_comp = converter_io.CHEM_COMPS[residue_name]

        chem_atoms_by_id = {
            chem_comp_atom.ID: chem_comp_atom for chem_comp_atom in chem_comp.chemAtoms
        }

        chem_atom_set_atoms = {}
        for chem_atom_set in chem_comp.chemAtomSets:
            chem_atom_set_atoms[(chem_atom_set.name, chem_atom_set.subType)] = (
                _get_chem_atom_set_atoms(chem_atom_set, chem_comp)
            )

        active_chem_comp_vars = [
            chem_comp_var
            for chem_comp_var in chem_comp.chemCompVars
            if chem_comp_var.isDefaultVar
            and chem_comp_var.linking in ["start", "middle", "end"]
        ]

        active_chem_comp_vars_atom_sets = {}
        for chem_comp_var in active_chem_comp_vars:
            active_chem_comp_atom_ids = set(chem_comp_var.chemAtoms)
            for atom_set_name, atom_set_atom_ids in chem_atom_set_atoms.items():
                if atom_set_atom_ids.issubset(active_chem_comp_atom_ids):
                    atom_set_chem_atoms = [
                        chem_atoms_by_id[id].name for id in atom_set_atom_ids
                    ]
                    name = atom_set_name[0]
                    name = (
                        name if name != "HD*|HE*" else "QR"
                    )  # i don't think NEF accepts HD*|HE* as a name
                    name = (
                        "f{name[:-1}%" if name[-1] == "*" else name
                    )  # NEF prioritises % over *
                    active_chem_comp_vars_atom_sets.setdefault(
                        chem_comp_var.linking.value, []
                    ).append((name, atom_set_chem_atoms))

        equivalent_atoms_by_residue[residue_name] = active_chem_comp_vars_atom_sets

    return equivalent_atoms_by_residue


LINKING_BY_NAME = {
    "free": ChemCompLinking.free,
    "start": ChemCompLinking.start,
    "middle": ChemCompLinking.middle,
    "end": ChemCompLinking.end,
}
NAME_BY_LINKING = {linking: name for name, linking in LINKING_BY_NAME.items()}


class BmrbShiftAmbiguities(Enum):
    UNIQUE = 1
    ALIPHATIC_GEMINAL = 2
    AROMATIC_GEMINAL = 3
    INTRA_RESIDUE = 4
    INTER_RESIDUE = 5
    INTER_MOLECULAR = 6
    UNKNOWN = 9


def rmsd(values):
    mean = sum(values) / len(values)
    squared_values = [(value - mean) ** 2 for value in values]
    mean_of_squared_values = sum(squared_values) / len(squared_values)
    return mean_of_squared_values**0.5


def pipe(
    nef_entry: Entry,
    chain_codes: List[str],
    frame_name: str,
    nmrstar_entry: Entry,
    file_name: Path,
    use_author: bool,
    stereo_mode: StereoAssignmentHandling,
):

    sequence_residues = sequence_from_entry_or_exit(nef_entry)

    sequence_residues_by_residue = {
        Residue.from_sequence_residue(sequence_residue): sequence_residue
        for sequence_residue in sequence_residues
    }

    residues = set(sequence_residues_by_residue.keys())

    denormalised_shifts, ambiguities, name_info = _chemical_shifts_from_star_frame(
        nmrstar_entry, use_author, file_name
    )

    stereo_mode = _get_stereo_mode(ambiguities, stereo_mode)

    entity_id_to_chain_code = _build_entity_ids_to_chain_codes(
        chain_codes, denormalised_shifts
    )

    denormalised_shifts = _replace_shift_entity_ids_with_chain_codes(
        denormalised_shifts, entity_id_to_chain_code
    )

    per_residue_and_atom_shifts = _organise_shifts_by_residue_and_atom_names(
        denormalised_shifts
    )

    shift_residues = set(per_residue_and_atom_shifts.keys())

    _exit_error_if_missing_residues(residues, shift_residues, file_name)

    residue_names = sorted(
        list(set([residue.residue_name for residue in per_residue_and_atom_shifts]))
    )

    ambiguities = _replace_atom_dict_entity_ids_with_chain_codes(
        ambiguities, entity_id_to_chain_code
    )

    per_residue_ambiguities = _organise_ambiguities_by_residue(ambiguities)

    _replace_atoms_with_duplicated_shifts(
        per_residue_and_atom_shifts,
        sequence_residues_by_residue,
        per_residue_ambiguities,
    )

    stereo_mode = StereoAssignmentHandling.ALL_AMBIGUOUS

    all_stereo_pairs = _get_geminal_pairs(residue_names)

    _apply_stereo_assignment_ambiguities(
        per_residue_and_atom_shifts,
        all_stereo_pairs,
        sequence_residues_by_residue,
        per_residue_ambiguities,
        stereo_mode,
    )

    all_shifts = sorted(
        flatten(
            [
                list(residue_shifts.values())
                for residue_shifts in per_residue_and_atom_shifts.values()
            ]
        )
    )

    if not frame_name:
        frame_name = name_info.entry_id + "__" + str(name_info.identifier)

    shifts_frame = shifts_to_nef_frame(ShiftList(all_shifts), frame_name)

    return add_frames_to_entry(
        nef_entry,
        [
            shifts_frame,
        ],
    )


def _apply_stereo_assignment_ambiguities(
    per_residue_and_atom_shifts,
    all_stereo_pairs,
    sequence_residues_by_residue,
    per_residue_ambiguities,
    stereo_mode,
):
    for residue, chemical_shifts_by_atom_names in per_residue_and_atom_shifts.items():
        sequence_residue = sequence_residues_by_residue[residue]
        geminal_pairs_for_residue = all_stereo_pairs[sequence_residue.residue_name]
        linking = sequence_residue.linking.lower()
        ambiguous_pairs = geminal_pairs_for_residue[linking]

        for ambiguous_pair in ambiguous_pairs.values():
            found_pairs = {}
            for before_atom_name in ambiguous_pair:
                if before_atom_name in chemical_shifts_by_atom_names:
                    found_pairs[before_atom_name] = chemical_shifts_by_atom_names[
                        before_atom_name
                    ]

            if len(found_pairs) == 2:
                found_pairs = dict(
                    sorted(found_pairs.items(), key=lambda item: item[1])
                )
                if stereo_mode is StereoAssignmentHandling.ALL_AMBIGUOUS:
                    for before, old_shift in found_pairs.items():
                        new_atom_name = ambiguous_pair[before]
                        # print('!', before, new)
                        atom = old_shift.atom
                        atom = replace(atom, atom_name=new_atom_name)
                        new_shift = replace(old_shift, atom=atom)
                        del chemical_shifts_by_atom_names[before]
                        chemical_shifts_by_atom_names[new_atom_name] = new_shift


def _replace_atoms_with_duplicated_shifts(
    per_residue_and_atom_shifts, sequence_residues_by_residue, per_residue_ambiguities
):

    residue_names = set(
        [residue.residue_name for residue in per_residue_and_atom_shifts.keys()]
    )

    atom_sets_by_residue = _get_atom_sets_by_residue(residue_names)
    for residue, chemical_shifts_by_atom_names in per_residue_and_atom_shifts.items():

        sequence_residue = sequence_residues_by_residue[residue]

        # ambiguity_keys = list(per_residue_ambiguities.keys())

        linking = sequence_residue.linking.value.lower()
        related_atom_sets = atom_sets_by_residue[sequence_residue.residue_name][linking]

        # active_atom_names = {shift.atom.atom_name for shift in chemical_shifts}

        # we do them by longest first so we get the most general match first
        related_atom_sets_by_length = reversed(
            sorted(related_atom_sets, key=lambda item: len(item[1]))
        )

        for atom_set_name, related_atoms in related_atom_sets_by_length:
            # print(residue, 'atom set name', atom_set_name, 'related atoms', related_atoms)
            shifts = [
                chemical_shifts_by_atom_names[atom]
                for atom in related_atoms
                if atom in chemical_shifts_by_atom_names
            ]

            if len(shifts) != len(related_atoms):
                continue

            new_ambiguities = set(
                [per_residue_ambiguities[residue][atom] for atom in related_atoms]
            )

            if len(new_ambiguities) != 1:
                new_ambiguity = max(new_ambiguities)

                msg = f"""
                WARNING: the ambiguities for the atoms in atom set {atom_set_name} in the atom set are not all the same
                i got {', '.join(new_ambiguities)} i chose {new_ambiguity} as the ambiguity for the atom set
                """

                print(msg, file=sys.stderr)
            else:
                new_ambiguity = new_ambiguities.pop()

            rmsd_val = rmsd([shift.value for shift in shifts])
            if rmsd_val < 1e-7:
                chemical_shifts_by_atom_names[atom_set_name] = shifts[0]
                for atom in related_atoms:
                    del chemical_shifts_by_atom_names[atom]

                residue_ambiguities = per_residue_ambiguities[residue]
                residue_ambiguities[atom_set_name] = new_ambiguity
                for atom in related_atoms:
                    del residue_ambiguities[atom]


def _get_stereo_mode(ambiguities, stereo_mode):
    if stereo_mode == StereoAssignmentHandling.AUTO and any(
        [
            value == BmrbShiftAmbiguities.ALIPHATIC_GEMINAL.value
            for value in ambiguities.values()
        ]
    ):
        stereo_mode = StereoAssignmentHandling.AS_ASSIGNED
    else:
        stereo_mode = StereoAssignmentHandling.ALL_AMBIGUOUS
    return stereo_mode


@dataclass
class ShiftListNameInfo:
    entry_id: str
    identifier: Union[int, str]


def _chemical_shifts_from_star_frame(nmrstar_entry, use_author, file_name):
    shift_lists = nmrstar_entry.get_saveframes_by_category("assigned_chemical_shifts")

    if len(shift_lists) == 0:
        msg = f"""\
        there are no saveframes of assigned chemical shifts in the file {file_name}
        """

        exit_error(msg)

    if len(shift_lists) != 1:
        msg = f"""\
        i can only cope with one saveframe of assigned chemical shifts currently there are {len(shift_lists)}
        in the file {file_name}
        """

        exit_error(msg)

    # can there be more than one? I guess yes
    ambiguities = {}
    denormalised_shifts = []

    frame_code = shift_lists[0].get_tag("Sf_framecode")[0]
    category = shift_lists[0].get_tag("Sf_category")[0]
    entry_id = shift_lists[0].get_tag("Entry_ID")[0]
    list_identifier = frame_code[len(category) + 1 :].lstrip("_")
    name_info = ShiftListNameInfo(entry_id, list_identifier)

    for i, shift_list in enumerate(shift_lists):

        try:
            shifts_loop = shift_list.get_loop("_Atom_chem_shift")

        except KeyError:
            msg = f"""\
                warning the saveframe {shift_list.get_tag('Sf_framecode')[0]}
                does not contain a loop with the tag _Atom_chem_shift
                """
            print(msg, file=sys.stderr)
            continue

        # TODO add warnings and check all required fields are present
        for row in loop_row_namespace_iter(shifts_loop):
            chain_code = row.Entity_assembly_ID
            if use_author:
                atom_id = row.Auth_atom_ID
                seq_id = row.Auth_seq_ID
                comp_id = row.Auth_comp_ID
            else:
                atom_id = row.Atom_ID
                seq_id = row.Seq_ID
                comp_id = row.Comp_ID

            if atom_id == UNUSED and use_author:
                atom_id = row.Atom_ID
            if seq_id == UNUSED and use_author:
                seq_id = row.Seq_ID
            if comp_id == UNUSED and use_author:
                comp_id = row.Comp_ID

            shift = row.Val
            sdev = row.Val_err
            element = row.Atom_type
            isotope = row.Atom_isotope_number
            ambiguity_code = row.Ambiguity_code

            residue = Residue(chain_code, seq_id, comp_id)
            atom = AtomLabel(residue, atom_id, element, isotope)

            shift = ShiftData(atom, shift, sdev)
            denormalised_shifts.append(shift)
            ambiguities[atom] = ambiguity_code

    return denormalised_shifts, ambiguities, name_info


def _exit_if_unknown_chem_comp(residue_name):
    import nef_pipelines.lib.translation.io as converter_io

    if not (residue_name) in converter_io.CHEM_COMPS:
        exit_error(
            f"the residue / molecule {residue_name} is not found in the chemical components dictionary"
        )


def _organise_shifts_by_residue_and_atom_names(denormalised_shifts):
    per_residue_shifts = {}
    for shift in denormalised_shifts:
        residue = shift.atom.residue
        per_residue_shifts.setdefault(residue, []).append(shift)
    per_residue_and_atom_shifts = {}
    for residue, shifts in per_residue_shifts.items():
        per_residue_and_atom_shifts[residue] = {
            shift.atom.atom_name: shift for shift in shifts
        }
    return per_residue_and_atom_shifts


def _organise_ambiguities_by_residue(ambiguities):
    # print(type(ambiguities))
    per_residue_ambiguities = {}
    for atom, ambiguity in ambiguities.items():
        residue = atom.residue
        # print(residue)
        per_residue_ambiguities.setdefault(residue, {})[atom.atom_name] = ambiguity
    return per_residue_ambiguities


def _replace_shift_entity_ids_with_chain_codes(
    denormalised_shifts, entity_id_to_chain_code
):
    for i, shift in enumerate(denormalised_shifts):
        atom = shift.atom
        residue = atom.residue
        residue = replace(
            residue, chain_code=entity_id_to_chain_code[residue.chain_code]
        )
        atom = replace(atom, residue=residue)
        denormalised_shifts[i] = replace(shift, atom=atom)

    return denormalised_shifts


def _replace_atom_dict_entity_ids_with_chain_codes(
    ambiguities, entity_id_to_chain_code
):
    result = {}
    for atom, ambiguity in ambiguities.items():
        residue = atom.residue
        residue = replace(
            residue, chain_code=entity_id_to_chain_code[residue.chain_code]
        )
        atom = replace(atom, residue=residue)
        result[atom] = ambiguity

    return result


def _build_entity_ids_to_chain_codes(chain_codes, denormalised_shifts):
    entity_ids = list(
        OrderedSet(
            sorted([shift.atom.residue.chain_code for shift in denormalised_shifts])
        )
    )
    entity_id_to_chain_code = {
        entity_id: chain_code
        for entity_id, chain_code in zip(entity_ids, chain_code_iter(chain_codes))
    }
    return entity_id_to_chain_code


def _exit_error_if_missing_residues(residues, shift_residues, file_name):

    # could do more here and see if there is a match by sequence...
    if not shift_residues.issubset(residues):
        msg = """\
            warning the residues in the shift list from {file_name} are not all present in the sequence you read
            from the NEF entry, the missing residues are
            {missing_residue_table}
            """
        msg = dedent(msg)
        missing_residues = sorted(shift_residues - residues)
        residue_names = [
            f"@{residue.chain_code}#{residue.sequence_code}.{residue.residue_name}"
            for residue in missing_residues
        ]
        residue_name_table = chunks(residue_names, 10)
        missing_residue_table = tabulate(  # noqa: F841
            residue_name_table, tablefmt="plain"
        )
        msg = f(msg)
        exit_error(msg)



from dataclasses import replace
from pathlib import Path
from typing import List

import typer
from ordered_set import OrderedSet
from pynmrstar import Entry

from nef_pipelines.lib.nef_lib import (
    UNUSED,
    add_frames_to_entry,
    loop_row_namespace_iter,
    read_entry_from_file_or_exit_error,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.sequence_lib import (
    chain_code_iter,
    offset_chain_residues,
    sequence_to_nef_frame,
)
from nef_pipelines.lib.structures import Residue
from nef_pipelines.lib.util import STDIN, exit_error, parse_comma_separated_options
from nef_pipelines.transcoders.fasta.importers.sequence import (
    _get_sequence_offsets,
    residues_to_chain_codes,
)
from nef_pipelines.transcoders.nmrstar import import_app

app = typer.Typer()

NO_CHAIN_START_HELP = """don't include the start chain link type on a chain for the first residue [linkage will be
                         middle] for the named chains. Either use a comma joined list of chains [e.g. A,B] or call this
                         option multiple times to set chain starts for multiple chains"""
NO_CHAIN_END_HELP = """don't include the end chain link type on a chain for the last residue [linkage will be
                       middle] for the named chains. Either use a comma joined list of chains [e.g. A,B] or call this
                       option multiple times to set chain ends for multiple chains"""


# todo add comment to other sequences etc
@import_app.command()
def sequence(
    chain_codes: List[str] = typer.Option(
        [],
        "--chains",
        help="chain codes to use for the imported chains, can be a a comma separated list or can be called "
        "multiple times",
        metavar="<CHAIN-CODES>",
    ),
    starts: List[str] = typer.Option(
        [],
        "--starts",
        help="first residue number of sequences can be a comma separated list or ",
        metavar="<START>",
    ),
    use_author: bool = typer.Option(False, help="use author field for sequence codes"),
    no_chain_starts: List[str] = typer.Option(
        [], "--no-chain-start", help=NO_CHAIN_START_HELP
    ),
    no_chain_ends: List[str] = typer.Option(
        [], "--no-chain-end", help=NO_CHAIN_END_HELP
    ),
    # # TODO: unused inputs!
    entry_name: str = typer.Option(
        None, help="a name for the entry (defaults to the bmrb entry number)"
    ),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--in",
        help="file to read NEF data from [- is stdin; defaults is stdin]",
    ),
    file_path: Path = typer.Argument(
        ..., help="the file to read", metavar="<NMR-STAR-FILE>"
    ),
):
    """- convert NMR_STAR sequence to NEF [alpha]"""

    chain_codes = parse_comma_separated_options(chain_codes)
    if not chain_codes:
        chain_codes = ["A"]

    no_chain_starts = parse_comma_separated_options(no_chain_starts)
    if not no_chain_starts:
        no_chain_starts = [False]

    no_chain_ends = parse_comma_separated_options(no_chain_ends)
    if no_chain_ends:
        no_chain_ends = [False]

    starts = [int(elem) for elem in parse_comma_separated_options(starts)]
    if not starts:
        starts = [1]

    nef_entry = read_or_create_entry_exit_error_on_bad_file(input)

    nmrstar_entry = read_entry_from_file_or_exit_error(file_path)

    if entry_name is None:
        entry_name = f"nmrstar_{nmrstar_entry.entry_id}"

    if len(nef_entry.frame_list) == 0:
        nef_entry.entry_id = entry_name

    nef_entry = pipe(
        nef_entry,
        chain_codes,
        starts,
        no_chain_starts,
        no_chain_ends,
        entry_name,
        nmrstar_entry,
        file_path,
        use_author,
    )

    print(nef_entry)


def pipe(
    nef_entry: Entry,
    chain_codes: List[str],
    starts: List[str],
    no_chain_starts: List[str],
    no_chain_ends: List[str],
    entry_name: str,
    nmrstar_entry: Entry,
    file_path: Path,
    use_author: bool,
):

    nef_entry.name = entry_name

    entities = nmrstar_entry.get_saveframes_by_category("entity")

    num_entities = len(entities)
    if num_entities == 0:
        msg = f"""
            No sequences (entities) found in {file_path}.
        """

        exit_error(msg)
    elif num_entities > 1:
        msg = f"""
            More than one sequence (entity) found in {file_path} i can only currently cope with 1.
        """
        exit_error(msg)

    entity = entities[0]

    sequence_loop = entity.get_loop("_Entity_comp_index")

    sequence_residues = []
    entity_ids = OrderedSet()
    # TODO: add ability to skip unknown chem comps
    # TODO add warnings and check all required fileds are present
    for row in loop_row_namespace_iter(sequence_loop):
        if use_author:
            seq_code = row.Auth_seq_ID
        else:
            seq_code = row.ID

        if seq_code == UNUSED and row.Auth_seq_ID != UNUSED:
            seq_code = row.Auth_seq_ID

        residue_name = row.Comp_ID
        entity_id = row.Entity_ID

        entity_ids.add(entity_id)
        residue = Residue(entity_id, seq_code, residue_name)
        sequence_residues.append(residue)

    entity_id_to_chain_code = {
        entity_id: chain_code
        for entity_id, chain_code in zip(entity_ids, chain_code_iter(chain_codes))
    }

    for i, residue in enumerate(sequence_residues):
        sequence_residues[i] = replace(
            residue, chain_code=entity_id_to_chain_code[residue.chain_code]
        )

    read_chain_codes = residues_to_chain_codes(sequence_residues)

    offsets = _get_sequence_offsets(read_chain_codes, starts)

    sequence_residues = offset_chain_residues(sequence_residues, offsets)

    sequence_residues = sorted(sequence_residues)

    nmrstar_frame = sequence_to_nef_frame(
        sequence_residues, no_chain_starts, no_chain_ends
    )

    return add_frames_to_entry(
        nef_entry,
        [
            nmrstar_frame,
        ],
    )



import string

import pytest

from nef_pipelines.lib.util import strip_characters_left
from nef_pipelines.transcoders.sparky.sparky_lib import parse_single_assignment


def test_strip_characters_left_empty():

    test = ""

    stripped, remaining = strip_characters_left(test, string.digits)

    assert stripped == ""
    assert remaining == ""


def test_strip_characters_left_partial():

    test = "123abc"

    stripped, remaining = strip_characters_left(test, string.digits)

    assert stripped == "123"
    assert remaining == "abc"


def test_strip_characters_left_all():

    test = "123456"

    stripped, remaining = strip_characters_left(test, string.digits)

    assert stripped == "123456"
    assert remaining == ""


@pytest.mark.parametrize(
    """input,expected_residue_name, expected_residue_number, expected_atom_name, previous_residue_name,
       previous_residue_number""",
    [
        ("10N", "", "10", "N", None, None),
        ("HN", "", "", "HN", None, None),
        ("?", "", "", "", None, None),
        ("T17H7", "T", "17", "H7", None, None),
        ("?N", "", "", "N", None, None),
        ("G16H8", "G", "16", "H8", None, None),
        ('T17H2"', "T", "17", 'H2"', None, None),
        ("H1'", "", "", "H1'", None, None),
        ("10N", "", "10", "N", "A", "11"),
        ("HN", "A", "11", "HN", "A", "11"),
        ("?", "", "", "", "A", "11"),
        ("T17H7", "T", "17", "H7", "A", "11"),
        ("?N", "A", "11", "N", "A", "11"),
        ("G16H8", "G", "16", "H8", "A", "11"),
        ('T17H2"', "T", "17", 'H2"', "A", "11"),
        ("H1'", "A", "11", "H1'", "A", "11"),
        ("N25CA", "N", "25", "CA", None, None),
    ],
)
def test_split_assignment(
    input,
    expected_residue_name,
    expected_residue_number,
    expected_atom_name,
    previous_residue_name,
    previous_residue_number,
):
    if previous_residue_number or previous_residue_name:
        residue_number, residue_name, atom_name = parse_single_assignment(
            input,
            previous_residue_name=previous_residue_name,
            previous_residue_number=previous_residue_number,
        )
    else:
        residue_number, residue_name, atom_name = parse_single_assignment(input)

    assert residue_name == expected_residue_name
    assert residue_number == expected_residue_number
    assert atom_name == expected_atom_name



import string
import sys
from enum import auto
from pathlib import Path
from typing import List

import requests
import typer
from fyeah import f
from pynmrstar import Entry
from requests.exceptions import HTTPError
from strenum import LowercaseStrEnum

from nef_pipelines.lib.nef_lib import (
    read_entry_from_file_or_exit_error,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.sequence_lib import chain_code_iter
from nef_pipelines.lib.util import (
    STDIN,
    exit_error,
    is_int,
    parse_comma_separated_options,
)
from nef_pipelines.transcoders.nmrstar import import_app
from nef_pipelines.transcoders.nmrstar.importers.sequence import pipe as sequence_pipe
from nef_pipelines.transcoders.nmrstar.importers.shifts import pipe as shift_pipe
from nef_pipelines.transcoders.nmrstar.nmrstar_lib import StereoAssignmentHandling

app = typer.Typer()

NO_CHAIN_START_HELP = """\
don't include the start chain link type on a chain for the first residue [linkage will be
middle] for the named chains. Either use a comma joined list of chains [e.g. A,B] or call this
option multiple times to set chain starts for multiple chains
"""

NO_CHAIN_END_HELP = """\
don't include the end chain link type on a chain for the last residue [linkage will be
middle] for the named chains. Either use a comma joined list of chains [e.g. A,B] or call this
option multiple times to set chain ends for multiple chains
"""

STEREO_HELP = """\
    how to handle stereo assignments the choices are:
    - ambiguous: assume all stereo assignments are ambiguous
    - as-assigned: use the stereo assignments as they are in the file
    - auto: use as assigned if some geminal stereo assignments are present, otherwise assume all are ambiguous
"""

BMRB_URL_TEMPLATE = "https://bmrb.io/ftp/pub/bmrb/entry_directories/bmr{entry_number}/bmr{entry_number}_3.str"

FILE_PATH_HELP = """\
the file to read, if it is of the form bmr<NUMBER> or just a number  or appears to be a url
the program will attempt to fetch the entry from the bmrb or the web first before looking
for a file on disc unless this behaviour overridden by the --source option
"""


class EntrySource(LowercaseStrEnum):
    FILE = auto()
    WEB = auto()
    AUTO = auto()


@import_app.command()
def project(
    chain_codes: List[str] = typer.Option(
        [],
        "--chains",
        help="""chain codes to use for the imported chains, can be a comma separated list or can be called
                multiple times if no chain codes are provided they will be named A B C...""",
        metavar="<CHAIN-CODES>",
    ),
    no_chain_starts: List[str] = typer.Option(
        [], "--no-chain-start", help=NO_CHAIN_START_HELP
    ),
    no_chain_ends: List[str] = typer.Option(
        [], "--no-chain-end", help=NO_CHAIN_END_HELP
    ),
    source: EntrySource = typer.Option(
        EntrySource.AUTO, help="the source of the entry"
    ),
    use_author: bool = typer.Option(False, help="use author field for sequence codes"),
    stereo_mode: StereoAssignmentHandling = typer.Option("auto", help=STEREO_HELP),
    entry_name: str = typer.Option(
        None, help="a name for the entry (defaults to the bmrb entry number)"
    ),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--in",
        help="file to read NEF data from [- is stdin; defaults is stdin]",
    ),
    url_template: str = typer.Option(
        BMRB_URL_TEMPLATE,
        help=f"template for the bmrb url [default {BMRB_URL_TEMPLATE}]",
    ),
    file_path: str = typer.Argument(..., help=", metavar=<NMR-STAR-FILE>"),
):
    """-  convert as much as possible from an NMR-STAR file to NEF [shifts & sequences] [alpha]"""

    chain_codes = parse_comma_separated_options(chain_codes)
    if not chain_codes:
        chain_codes = chain_code_iter()

    no_chain_starts = parse_comma_separated_options(no_chain_starts)
    if not no_chain_starts:
        no_chain_starts = [False]

    no_chain_ends = parse_comma_separated_options(no_chain_ends)
    if no_chain_ends:
        no_chain_ends = [False]

    is_bmrb = False
    if source in (EntrySource.AUTO, EntrySource.WEB):
        url, is_bmrb = _get_path_as_url_or_none(file_path, url_template)

        exit_on_error = True if source == EntrySource.WEB else False
        possible_entry = _get_bmrb_entry_from_web_or_none(url, exit_on_error)

        nmrstar_entry = _parse_text_to_star_or_none(possible_entry)

    if (source == EntrySource.AUTO and not nmrstar_entry) or source == EntrySource.FILE:
        nmrstar_entry = read_entry_from_file_or_exit_error(file_path)

    if not nmrstar_entry:
        msg = f"could not read entry from {file_path}"
        exit_error(msg)

    entry_name = nmrstar_entry.entry_id if not entry_name else entry_name
    entry_name = f"bmr{entry_name}" if is_bmrb else entry_name

    nef_entry = read_or_create_entry_exit_error_on_bad_file(
        input, entry_name=entry_name
    )

    entry = pipe(
        nef_entry,
        nmrstar_entry,
        chain_codes,
        no_chain_starts,
        no_chain_ends,
        use_author,
        stereo_mode,
        file_path,
    )

    print(entry)


def pipe(
    nef_entry,
    nmrstar_entry,
    chain_codes,
    no_chain_starts,
    no_chain_ends,
    use_author,
    stereo_mode,
    file_path,
):

    # def pipe(
    #         nef_entry: Entry,
    #         chain_codes: List[str],
    #         starts: List[str],
    #         no_chain_starts: List[str],
    #         no_chain_ends: List[str],
    #         entry_name: str,
    #         nmrstar_entry: Entry,
    #         file_path: Path,
    #         use_author: bool,
    # ):

    entry_id = nmrstar_entry.entry_id
    nef_entry = sequence_pipe(
        nef_entry,
        chain_codes,
        [],
        no_chain_starts,
        no_chain_ends,
        entry_id,
        nmrstar_entry,
        file_path,
        use_author,
    )

    # def pipe(
    #         nef_entry: Entry,
    #         chain_codes: List[str],
    #         frame_name: str,
    #         nmrstar_entry: Entry,
    #         file_name: Path,
    #         use_author: bool,
    #         stereo_mode: StereoAssignmentHandling,
    # ):

    nef_entry = shift_pipe(
        nef_entry, [], entry_id, nmrstar_entry, file_path, use_author, stereo_mode
    )

    return nef_entry


def _parse_text_to_star_or_none(possible_entry):
    nmrstar_entry = None
    if possible_entry and possible_entry.startswith(b"data_"):
        try:
            nmrstar_entry = Entry.from_string(possible_entry.decode("utf-8"))
        except Exception:
            pass
    return nmrstar_entry


def _get_bmrb_entry_from_web_or_none(url, exit_on_error=True):
    possible_entry = None
    error = False
    exception = None
    if url:
        try:
            response = requests.get(url)
            response.raise_for_status()
            possible_entry = response.content
        except HTTPError as http_err:
            msg = f"while trying to download the url {url} there was an http error {http_err}"
            exception = http_err
        except Exception as err:
            msg = f"while trying to download the url {url} there was an error {err}"
            exception = err

    if error:
        if exit_on_error:
            exit_error(msg, exception)
        else:
            print(f"WARNING: {msg}, trying to read as a local file", file=sys.stderr)

    return possible_entry


def _get_path_as_url_or_none(file_path, url_template):
    url = None
    is_bmrb = False
    if file_path.startswith("https://"):
        url = file_path
    elif file_path.startswith("bmr"):
        entry_number = file_path[3:]
        entry_check = entry_number.lstrip(string.digits)
        if len(entry_check) == 0:
            url = f(url_template)
            is_bmrb = True
    elif is_int(file_path):
        entry_number = file_path
        url = f(url_template)
        is_bmrb = True

    return url, is_bmrb



import itertools
import string
import sys
from copy import copy
from dataclasses import replace
from datetime import time
from enum import auto
from itertools import product
from pathlib import Path
from random import seed as set_random_seed
from random import shuffle
from typing import Dict, List

import typer
from ordered_set import OrderedSet
from strenum import LowercaseStrEnum

from nef_pipelines.lib.nef_lib import (
    SELECTORS_LOWER,
    UNUSED,
    SelectionType,
    loop_row_dict_iter,
    read_or_create_entry_exit_error_on_bad_file,
    select_frames,
)
from nef_pipelines.lib.sequence_lib import atom_sort_key
from nef_pipelines.lib.structures import AtomLabel, Residue
from nef_pipelines.lib.util import (
    STDIN,
    chunks,
    exit_error,
    is_int,
    parse_comma_separated_options,
    strip_characters_left,
)
from nef_pipelines.tools.frames import frames_app

SELECTOR_HELP = """
    limit changes to a a particular frame by a selector which can be a frame name or category
    note: wildcards [*] are allowed. Frames are selected by name and subsequently by category if the name
    doesn't match [-t /--selector-type allows you to force which selection type to use]. If no frame
    names or categories are provided all frames are unassigned.
"""

ATOM_LABEL_FIELDS = "chain_code sequence_code residue_name atom_name".split()


class Targets(LowercaseStrEnum):
    CHAIN_CODE = auto()
    SEQUENCE_CODE = auto()
    RESIDUE_NAME = auto()
    ATOM_NAME = auto()
    OFFSETS = auto()
    ALL = auto()
    NOT_ATOM = auto()


ALL_COMPONENTS = {
    Targets.CHAIN_CODE,
    Targets.SEQUENCE_CODE,
    Targets.RESIDUE_NAME,
    Targets.ATOM_NAME,
}
NOT_ATOM_COMPONENTS = copy(ALL_COMPONENTS)
NOT_ATOM_COMPONENTS.remove(Targets.ATOM_NAME)

TARGET_LOOKUP = {
    str(Targets.CHAIN_CODE): {
        Targets.CHAIN_CODE,
    },
    str(Targets.SEQUENCE_CODE): {
        Targets.SEQUENCE_CODE,
    },
    str(Targets.RESIDUE_NAME): {
        Targets.RESIDUE_NAME,
    },
    str(Targets.ATOM_NAME): {
        Targets.ATOM_NAME,
    },
    str(Targets.ALL): ALL_COMPONENTS,
    str(Targets.NOT_ATOM): NOT_ATOM_COMPONENTS,
}

# TODO: allow unique abbreviations
TARGETS_HELP = f"""
    choose what to de-assign it can be a comma separated list or called multiple times. Possible values are:
    {', '.join([e.value for e in Targets])}.
    all and not_atom are combinations, all is: chain_code + sequence_code + residue_name + atom_name and
    not_atom is: chain_code + sequence_code + residue_name.
"""


# could do with an ordered mode which would sort assignments
# but would require a shift sorter
class SequenceMode(LowercaseStrEnum):
    UNUSED = auto()
    INDEX = auto()
    ORDERED = auto()
    RANDOM = auto()
    PRESERVE = auto()


# TODO: how do you decide whether its a selection within a chain or the file
SEQUENCE_MODE_HELP = f"""
    value to deassign the sequence code to [one of {', '.join([e.value for e in SequenceMode])}] UNUSED is .,
    INDEX is the index of the assignment in its chain, RANDOM is a random index for the assignment
    within its chain, ORDERED retains the order of the residues when deassigned to pseudo residues, PRESERVE deassigns
    residues to pseudo residues with the same number as the assigned residue any clashes are disambiguated by renaming
    existing pseudo residues.
"""

CHAIN_VALUE_HELP = """
    value to deassign chains to [note @- is the default nmr chain and . is no chain] if more than one value is specified
    [using multiple options or a comma separated list] they will be treated as old and new value pairs, with old values
    being reassigned to the new values. Current values which are not in the new values will be left untouched.
"""

RANDOM_SEED_HELP = """
    set the random seed to an integer, any value less than zero will set the seed from the current clock and
    note the value
"""
ANY = "*"

SELECTION_TYPE_HELP = f"""
    how to select frames to renumber, can be one of: {SELECTORS_LOWER}.
    Any will match on frame name first and then if there is no match attempt to match on category
"""

# TODO: add the ability to unassign a particular dimension
# TODO: add the ability specify the pseudo residue prefix


@frames_app.command()
def unassign(
    selector_type: SelectionType = typer.Option(
        SelectionType.ANY, "-t", "--selector-type", help=SELECTION_TYPE_HELP
    ),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--input",
        metavar="|PIPE|",
        help="input to read NEF data from [- is stdin]",
    ),
    targets: List[Targets] = typer.Option(
        [Targets.NOT_ATOM], "-t", "--targets", help=TARGETS_HELP
    ),
    chain_values: List[str] = typer.Option(
        [], metavar="<CHAIN> | <OLD-CHAIN>: <NEW-CHAIN>...", help=CHAIN_VALUE_HELP
    ),
    sequence_mode: SequenceMode = typer.Option(
        SequenceMode.INDEX, metavar="<MODE>", help=SEQUENCE_MODE_HELP
    ),
    random_seed: int = typer.Option(42, metavar="<RANDOM-SEED>", help=RANDOM_SEED_HELP),
    frame_selectors: List[str] = typer.Argument(
        None,
        help=SELECTOR_HELP,
    ),
):
    """- unassign frames in the current input"""

    entry = read_or_create_entry_exit_error_on_bad_file(input)

    if random_seed < 0:
        seed = int(time.time())
        set_random_seed(seed)
        print(f"SEED is set to: {seed}", file=sys.stderr)
    else:
        set_random_seed(random_seed)

    chain_values = parse_comma_separated_options(chain_values)
    num_chain_values = len(chain_values)
    if num_chain_values == 0:
        chain_mappings = {ANY: "@-"}
    elif num_chain_values > 1 and num_chain_values % 2 != 0:
        msg = f"""
            if there are more than two new chain_code values specified they must come in pairs i got
            {num_chain_values}, the values were {', '.join(chain_values)}
        """
        exit_error(msg)

    elif num_chain_values == 1:
        chain_mappings = {ANY: chain_values[0]}
    else:
        chain_mappings = {old: new for old, new in chunks(chain_values, 2)}

    raw_targets = parse_comma_separated_options(targets)

    targets = _build_targets(raw_targets)

    if Targets.CHAIN_CODE in targets and len(chain_mappings) == 0:
        chain_mappings_msg = ", ".join(
            [f"{first} -> {second}" for first, second in chain_mappings.items()]
        )

        msg = f"""\
            you can't have chain mappings and not target chains, you might want to add CHAIN_CODE to --targets!
            targets were: {', '.join([str(target) for target in targets])}
            chain mappings were: {chain_mappings_msg}
        """

        exit_error(msg)

    frame_selectors = {
        (frame.category, frame.name)
        for frame in select_frames(entry, selector_type, frame_selectors)
    }

    entry = pipe(entry, frame_selectors, targets, sequence_mode, chain_mappings)

    print(entry)


# noinspection PyUnusedLocal
def _build_targets(raw_targets: List[Targets]) -> List[Targets]:
    result = set()

    for raw_target in raw_targets:
        result.update(TARGET_LOOKUP[raw_target])

    return result


def pipe(entry, frame_selectors, targets, sequence_mode, chain_mappings):

    current_assignments = OrderedSet()
    for frame in entry:
        if (frame.category, frame.name) in frame_selectors:
            for loop in frame:
                for i, row in enumerate(loop_row_dict_iter(loop)):
                    assignments_by_index = {}
                    for index, prefix in product(range(1, 16), ATOM_LABEL_FIELDS):

                        if prefix in row:
                            key = prefix
                        else:
                            key = f"{prefix}_{index}"

                            if key not in row:
                                break

                        assignments_by_index.setdefault(index, {})[prefix] = row[key]

                    for assignment in assignments_by_index.values():
                        label = _dict_to_label(assignment)

                        if label:
                            current_assignments.add(label)

    if sequence_mode == SequenceMode.ORDERED:
        current_assignments = sorted(current_assignments, key=atom_sort_key)

    assignment_map = _map_assignments(
        current_assignments, targets, sequence_mode, chain_mappings
    )

    for frame in entry:
        if (frame.category, frame.name) in frame_selectors:
            for loop in frame:
                tag_indices = {tag: index for index, tag in enumerate(loop.tags)}
                for row_index, row in enumerate(loop_row_dict_iter(loop)):
                    assignments_by_index = {}
                    for index, prefix in product(range(1, 16), ATOM_LABEL_FIELDS):

                        if prefix in row:
                            key = prefix
                        else:
                            key = f"{prefix}_{index}"

                        if key not in row:
                            break

                        assignments_by_index.setdefault(index, {})[prefix] = row[key]

                    for column_index, assignment in enumerate(
                        assignments_by_index.values(), start=1
                    ):
                        old_label = _dict_to_label(assignment)

                        if old_label is not None:

                            new_label = assignment_map[old_label]

                            for atom_label_field in ATOM_LABEL_FIELDS:
                                key = None
                                if atom_label_field in tag_indices:
                                    key = atom_label_field
                                if not key:
                                    possible_key = f"{atom_label_field}_{column_index}"
                                    if possible_key in tag_indices:
                                        key = possible_key

                                if key:
                                    tag_index = tag_indices[key]

                                    if atom_label_field == "chain_code":
                                        new_value = new_label.residue.chain_code
                                    elif atom_label_field == "sequence_code":
                                        new_value = new_label.residue.sequence_code
                                    elif atom_label_field == "residue_name":
                                        new_value = new_label.residue.residue_name
                                    elif atom_label_field == "atom_name":
                                        new_value = new_label.atom_name

                                    loop.data[row_index][tag_index] = new_value

    return entry


def _dict_to_label(assignment):
    result = None
    if "atom_name" in assignment:
        atom_name = assignment["atom_name"]
        residue_values = {
            key: assignment[key]
            for key in assignment
            if key != "atom_name" and key in ATOM_LABEL_FIELDS
        }
        residue = Residue(**residue_values)
        result = AtomLabel(residue, atom_name)
    return result


def _map_assignments(
    assignments: List[AtomLabel], targets, sequence_mode, chain_mappings
) -> Dict[AtomLabel, AtomLabel]:

    assignment_map = {}
    if Targets.CHAIN_CODE in targets:
        for assignment in assignments:

            if sequence_mode == SequenceMode.UNUSED:
                new_chain_code = "."
            elif len(chain_mappings) == 1 and list(chain_mappings.keys())[0] == ANY:
                new_chain_code = list(chain_mappings.values())[0]
            else:
                target_chain_code = assignment.residue.chain_code
                new_chain_code = target_chain_code
                if assignment.residue.chain_code in chain_mappings:
                    new_chain_code = chain_mappings[target_chain_code]

            new_residue = replace(assignment.residue, chain_code=new_chain_code)
            new_assignment = replace(assignment, residue=new_residue)

            assignment_map[assignment] = new_assignment

    if Targets.RESIDUE_NAME in targets:
        for assignment, new_assignment in assignment_map.items():

            new_residue = replace(new_assignment.residue, residue_name=UNUSED)
            new_assignment = replace(new_assignment, residue=new_residue)

            assignment_map[assignment] = new_assignment

    if Targets.ATOM_NAME in targets:
        for assignment, new_assignment in assignment_map.items():

            new_assignment = replace(new_assignment, atom_name=UNUSED)

            assignment_map[assignment] = new_assignment

    if Targets.SEQUENCE_CODE in targets:
        new_assignment_by_chain = {}
        for assignment, new_assignment in assignment_map.items():
            chain_code = new_assignment.residue.chain_code
            new_assignment_by_chain.setdefault(chain_code, []).append(
                (assignment, new_assignment)
            )

        sequence_code_map = {}

        if len(new_assignment_by_chain) > 1:
            msg = """
                !WARNING! there is more than one chain present, this code hasn't been tested with multiple chains...
            """
            print(msg, file=sys.stderr)

        for chain, assignments in new_assignment_by_chain.items():
            old_sequence_codes = OrderedSet()
            for _, new_assignment in assignments:

                sequence_code = str(new_assignment.residue.sequence_code)
                if "-" in sequence_code:
                    sequence_code = "-".join(sequence_code.split("-")[:-1])
                old_sequence_codes.add(sequence_code)

            if sequence_mode == SequenceMode.UNUSED:
                for old_sequence_code in old_sequence_codes:
                    sequence_code_map[old_sequence_code] = UNUSED
            elif sequence_mode in (
                SequenceMode.INDEX,
                SequenceMode.RANDOM,
                SequenceMode.ORDERED,
            ):
                # pull out unassigned sequence_codes
                used_pseudo_residue_sequence_codes = [
                    sequence_code
                    for sequence_code in old_sequence_codes
                    if sequence_code[0] not in string.digits
                ]
                old_sequence_codes = [
                    sequence_code
                    for sequence_code in old_sequence_codes
                    if sequence_code[0] in string.digits
                ]

                new_sequence_code_counter = itertools.count(1)
                new_sequence_codes = []
                for _ in old_sequence_codes:

                    putative_sequence_code = next(new_sequence_code_counter)
                    while (
                        f"@{putative_sequence_code}"
                        in used_pseudo_residue_sequence_codes
                    ):
                        putative_sequence_code = next(new_sequence_code_counter)

                    new_sequence_codes.append(f"@{putative_sequence_code}")

                if sequence_mode == SequenceMode.RANDOM:
                    shuffle(new_sequence_codes)

                for i, old_sequence_code in enumerate(old_sequence_codes):
                    sequence_code_map[old_sequence_code] = new_sequence_codes[i]

                for sequence_code in used_pseudo_residue_sequence_codes:
                    sequence_code_map[sequence_code] = sequence_code

            elif sequence_mode == SequenceMode.PRESERVE:

                # TODO: note there is a subtle difference here to above can this be removed
                used_pseudo_residue_sequence_codes = [
                    sequence_code
                    for sequence_code in old_sequence_codes
                    if not is_int(sequence_code)
                ]
                old_sequence_codes = [
                    sequence_code
                    for sequence_code in old_sequence_codes
                    if is_int(sequence_code)
                ]

                sequence_code_map = {}

                for sequence_code in old_sequence_codes:
                    sequence_code_map[sequence_code] = f"@{sequence_code}"

                possible_new_sequence_codes = {}
                for used_sequence_code in used_pseudo_residue_sequence_codes:
                    if used_sequence_code == UNUSED:
                        possible_new_sequence_codes[UNUSED] = UNUSED
                    else:
                        for i in itertools.chain(
                            [
                                "",
                            ],
                            itertools.count(1),
                        ):

                            prefix, sequence_code = strip_characters_left(
                                used_sequence_code, string.ascii_letters + "@#"
                            )

                            new_sequence_code = f"{prefix}PR_{i}{sequence_code}"
                            if new_sequence_code in sequence_code_map:
                                continue

                            possible_new_sequence_codes[
                                used_sequence_code
                            ] = new_sequence_code
                            break

                sequence_code_map.update(possible_new_sequence_codes)

        for assignment, new_assignment in assignment_map.items():

            old_sequence_code = str(new_assignment.residue.sequence_code)
            offset = None
            if "-" in old_sequence_code:
                fields = old_sequence_code.split("-")
                if is_int(fields[-1]):
                    offset = fields[-1]
                    old_sequence_code = "-".join(fields[:-1])

            new_sequence_code = sequence_code_map[old_sequence_code]
            if sequence_mode == SequenceMode.UNUSED:
                new_sequence_code = UNUSED
            elif offset is not None:
                new_sequence_code = f"{new_sequence_code}-{offset}"

            updated_residue = replace(
                new_assignment.residue, sequence_code=new_sequence_code
            )

            if updated_residue.sequence_code == UNUSED:
                updated_residue = replace(updated_residue, chain_code=UNUSED)

            updated_atom_label = replace(new_assignment, residue=updated_residue)

            assignment_map[assignment] = updated_atom_label

    return assignment_map



import typer

import nef_pipelines
from nef_pipelines import nef_app

app = typer.Typer()
import_app = typer.Typer()

if nef_app.app:
    nef_app.app.add_typer(
        app, name="nmrpipe", help="-  read nmrpipe [peaks shifts & sequencess]"
    )

    app.add_typer(
        import_app, name="import", help="-  import nmrpipe [peaks, shifts & sequences]"
    )

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.transcoders.nmrpipe.importers.peaks  # noqa: F401
    import nef_pipelines.transcoders.nmrpipe.importers.sequence  # noqa: F401
    import nef_pipelines.transcoders.nmrpipe.importers.shifts  # noqa: F401



from itertools import tee
from pathlib import Path
from typing import List

import typer
from typer import Argument, Option

from nef_pipelines.lib.nef_lib import (
    NEF_MOLECULAR_SYSTEM,
    SELECTORS_LOWER,
    SelectionType,
    read_entry_from_file_or_stdin_or_exit_error,
    select_frames,
)
from nef_pipelines.lib.sequence_lib import sequence_from_frame
from nef_pipelines.lib.util import chunks, end_with_ordinal, exit_error, is_int
from nef_pipelines.tools.chains import chains_app

app = typer.Typer()

CHAIN_CODE = "chain_code"

SEQUENCE_CODE = "sequence_code"


# noinspection PyUnusedLocal
@chains_app.command()
def renumber(
    chain_offsets: List[str] = Argument(
        None,
        metavar="<CHAIN-CODE> <OFFSET/START>",
        help="chain-codes and offsets/starts for renumbering residue numbers, offsets are used by default and chain "
        "starts are available using the --starts option",
    ),
    input: Path = typer.Option(
        Path("-"),
        "-i",
        "--in",
        metavar="NEF-FILE",
        help="where to read NEF data from either a file or stdin '-'",
    ),
    selector_type: SelectionType = typer.Option(
        SelectionType.ANY,
        "-s",
        "--selector",
        help=f"control how to select frames to renumber, can be one of: {SELECTORS_LOWER}. "
        "Any will match on names first and then if there is no match attempt to match on category",
    ),
    starts: bool = Option(
        False,
        "-s",
        "--starts",
        help="renumber residues using a starting value for the chain "
        "rather than an offset",
    ),
    frame_selectors: List[str] = Option(
        None,
        "-f",
        "--frames",
        metavar="<FRAME-SELECTOR>",
        help="Limit changes to a particular frame by name or category [the selector], note: wildcards are "
        "allowed. Frames are selected by name and then category if the name doesn't match"
        " [the option -t /--selector-type allows you to choose which selection type to use]. If no frame"
        "selectors are provided all frames are renumbered",
    ),
):
    """- renumber chains in a nef file"""

    chain_offsets = get_chain_offset_pairs_or_exit(chain_offsets)

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    if starts:
        molecular_system_frames = entry.get_saveframes_by_category(NEF_MOLECULAR_SYSTEM)

        exit_if_molecular_system_isnt_a_singleton(molecular_system_frames)

        molecular_system = molecular_system_frames[0]

        chain_offsets = chain_starts_to_offsets(molecular_system, chain_offsets)

    frames = select_frames(entry, selector_type, frame_selectors)

    offset_chains_in_frames(frames, chain_offsets, starts)

    print(entry)

    # # get_help()
    #
    # # typer_click_object = typer.main.get_command(chains_app)
    # # print(chains_app.typer_instance)
    # print(typer_click_object.command()())


def chain_starts_to_offsets(molecular_system, chain_offsets):
    result = []
    for chain, start in chain_offsets:
        sorted_residues = sorted(sequence_from_frame(molecular_system, chain))

        for residue in sorted_residues:

            if is_int(residue.sequence_code):
                start_sequence_code = int(sorted_residues[0].sequence_code)
                break

        result.append([chain, start - start_sequence_code])

    return result


# def chain_starts_to_offsets(molecular_system, chain_offsets):
#     result = []
#     for chain, start in chain_offsets:
#         sorted_residues = sorted(sequence_from_frame(molecular_system, chain))
#
#         if len(sorted_residues) > 0:
#             start_sequence_code = int(sorted_residues[0].sequence_code)
#
#             result.append([chain, start - start_sequence_code])
#
#     return result


def exit_if_molecular_system_isnt_a_singleton(molecular_system_frames):
    number_of_molecular_system_frames = len(molecular_system_frames)
    if number_of_molecular_system_frames == 0:
        exit_error("there must be a molecular system defined to use the starts option")
    elif number_of_molecular_system_frames > 1:
        # could go ahead and combine but warn of inconsistent results
        exit_error(
            "there can only be one  molecular system frame defined in a NEF file"
        )


def offset_chains_in_frames(frames, chain_offsets, starts):
    for frame in frames:
        for chain, offset in chain_offsets:
            offset_residue_numbers(frame, chain, offset)


def get_chain_offset_pairs_or_exit(chain_offsets):

    if not chain_offsets:
        exit_error("you didn't provide any chains and offsets/starts")

    # tee here as we use the chain_offsets generator twice
    chain_offsets_check, chain_offsets = tee(chunks(chain_offsets, 2))

    for chain_offset in chain_offsets_check:
        if len(chain_offset) % 2:
            chain_offset_pairs = []

            for chain_offset in chain_offsets:
                chain_offset_pairs.append(":".join(chain_offset))

            msg = f"""\
                there must be an offset/start for each chain
                i got the following pairs of chain:offset/start: {', '.join(chain_offset_pairs)}
                the last one is missing an offset.
            """
            exit_error(msg)

    result = []
    for i, (chain, offset) in enumerate(chain_offsets, start=1):

        exit_if_offset_isnt_int(chain, offset, i)

        offset = int(offset)

        result.append((chain, offset))

    return result


def exit_if_offset_isnt_int(chain, offset, i):
    if not is_int(offset):
        msg = f"""\
                the offset/start {offset} in the {end_with_ordinal(i)} chain offset/start pair: {chain} {offset}
                can't be converted to an int
            """
        exit_error(msg)


def exit_if_chains_and_offsets_dont_match(chains, offsets):
    num_chains = len(chains)
    num_offsets = len(offsets)
    if num_chains != num_offsets:
        msg = f"""\
            chains and offsets must have the same length i got {num_chains} chains and {num_offsets}
            chains: {','.join(chains)}
            offsets: {','.join([str(offset) for offset in offsets])}
        """
        exit_error(msg)


def offset_residue_numbers(frame, chain, offset, starts=False):

    for loop_data in frame.loop_dict.values():
        for tag in loop_data.tags:
            if tag_is_based_on(tag, SEQUENCE_CODE):
                sequence_index = loop_data.tag_index(tag)

                chain_code_tag = tag_based_on(tag, SEQUENCE_CODE, CHAIN_CODE)
                chain_code_index = loop_data.tag_index(chain_code_tag)
                for line in loop_data:
                    if chain_code_index is not None and line[chain_code_index] == chain:
                        sequence = line[sequence_index]
                        if is_int(sequence):
                            sequence = int(sequence)
                            sequence = sequence + offset
                            line[sequence_index] = str(sequence)


def tag_based_on(tag, base, substitute):
    column_code = tag[len(base) :]
    new_tag = f"{substitute}{column_code}"
    return new_tag


def tag_is_based_on(tag, base):
    return tag.startswith(base)


#
#
# def read_args():
#     global parser
#     parser = argparse.ArgumentParser(
#         description="Add chemical shift errors to a NEF file"
#     )
#     parser.add_argument(
#         "-c",
#         "--chain",
#         metavar="CHAIN",
#         type=str,
#         default="A",
#         dest="chain",
#         help="chain in which to offset residue numbers",
#     )
#     parser.add_argument(
#         "-o",
#         "--offset",
#         metavar="OFFSET",
#         type=int,
#         default=0,
#         dest="offset",
#         help="offset to add to residue numbers",
#     )
#     parser.add_argument(metavar="FILES", nargs=argparse.REMAINDER, dest="files")
#
#     return parser.parse_args()
#
#
# def check_stream():
#     # make stdin a non-blocking file
#     fd = sys.stdin.fileno()
#     fl = fcntl.fcntl(fd, fcntl.F_GETFL)
#     fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)
#
#     try:
#         result = sys.stdin.read()
#     except Exception:
#         result = ""
#
#     return result
#
#
# if __name__ == "__main__":
#
#     args = read_args()
#
#     is_tty = sys.stdin.isatty()
#
#     if is_tty and len(args.files) == 0:
#         parser.print_help()
#         print()
#         msg = """I require at least 1 argument or input stream with a chemical_shift_list frame"""
#
#         exit_error(msg)
#
#     entries = []
#     try:
#         if len(args.files) == 0:
#             lines = check_stream()
#             if len(lines) != 0:
#                 entries.append(Entry.from_string(lines))
#             else:
#                 exit_error("Error: input appears to be empty")
#         else:
#             for file in args.files:
#                 entries.append(Entry.from_file(file))
#     except OSError as e:
#         msg = f"couldn't open target nef file because {e}"
#         exit_error(msg)
#
#     for entry in entries:
#         offset_residue_numbers(entry, args.chain, args.offset)
#
#         # print(entry)



import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_frame,
    isolate_loop,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.echidna.importers.peaks import peaks

app = typer.Typer()
app.command()(peaks)


EXPECTED = """
loop_
   _nef_peak.index
   _nef_peak.peak_id
   _nef_peak.chain_code_1
   _nef_peak.sequence_code_1
   _nef_peak.residue_name_1
   _nef_peak.atom_name_1
   _nef_peak.chain_code_2
   _nef_peak.sequence_code_2
   _nef_peak.residue_name_2
   _nef_peak.atom_name_2
   _nef_peak.position_1
   _nef_peak.position_uncertainty_1
   _nef_peak.position_2
   _nef_peak.position_uncertainty_2
   _nef_peak.height
   _nef_peak.height_uncertainty
   _nef_peak.volume
   _nef_peak.volume_uncertainty
   _nef_peak.ccpn_figure_of_merit
   _nef_peak.ccpn_comment

  0    0    A   10   ALA   N   A   10   ALA   H   122.836   .   7.826   .   .   .   .   .   6.03818   'large violation!'
  1    1    A   11   ALA   N   A   11   ALA   H   118.678   .   8.063   .   .   .   .   .   6.36477   'large violation!'
  2    2    A   12   ALA   N   A   12   ALA   H   119.301   .   7.842   .   .   .   .   .   4.6192    'large violation!'
  3    3    A   13   ALA   N   A   13   ALA   H   122.946   .   8.232   .   .   .   .   .   3.63582   'large violation!'
  4    4    A   14   ALA   N   A   14   ALA   H   120.475   .   8.948   .   .   .   .   .   3.07293   .
  5    5    A   15   ALA   N   A   15   ALA   H   118.138   .   8.798   .   .   .   .   .   2.89933   .
  6    6    A   16   ALA   N   A   16   ALA   H   120.784   .   8.035   .   .   .   .   .   3.00885   .
stop_
"""

EXPECTED_TENSOR = """
save_np_tensor_frame_echidna_peaks
   _np_tensor_frame.sf_category          np_tensor_frame
   _np_tensor_frame.sf_framecode         np_tensor_frame_echidna_peaks
   _np_tensor_frame.restraint_origin     measured
   _np_tensor_frame.ccpn_format          angles_euler
   _np_tensor_frame.restraint_magnitude  -3012.1
   _np_tensor_frame.restraint_rhomicity  -1797.8
   _np_tensor_frame.ccpn_phi             80.68
   _np_tensor_frame.ccpn_psi             164.6
   _np_tensor_frame.ccpn_theta           128.22
save_
"""

EXPECTED_POSITION = """
save_np_atom_position_echidna_peaks
   _np_atom_position.sf_category   np_atom_position
   _np_atom_position.sf_framecode  np_atom_position_echidna_peaks
   _np_atom_position.x             80.68
   _np_atom_position.y             17.426
   _np_atom_position.z             7.735

save_


"""


def test_basic():

    path = path_in_test_data(__file__, "echidna_peaks.txt")

    data_sequence = open(path_in_test_data(__file__, "echidna_sequence.nef")).read()

    result = run_and_report(app, [path], input=data_sequence)

    loop = isolate_loop(result.stdout, "nef_nmr_spectrum_echidna_peaks", "nef_peak")

    assert_lines_match(EXPECTED, str(loop))

    frame = isolate_frame(result.stdout, "nef_nmr_spectrum_echidna_peaks")

    frame = " ".join(frame.split())
    assert (
        "_nef_nmr_spectrum.ccpn_comment 'NOTE: merits with values > 3.63582 are flagged as large violations"
        in frame
    )
    assert (
        "_nef_nmr_spectrum.np_tensor_frame_name np_tensor_frame_echidna_peaks" in frame
    )
    assert (
        "_nef_nmr_spectrum.np_atom_position_name np_atom_position_echidna_peaks"
        in frame
    )

    tensor_frame = isolate_frame(result.stdout, "np_tensor_frame_echidna_peaks")

    assert_lines_match(tensor_frame, EXPECTED_TENSOR)

    position_frame = isolate_frame(result.stdout, "np_atom_position_echidna_peaks")

    assert_lines_match(position_frame, EXPECTED_POSITION)
